from typing import Match

from loguru import logger

import tf2mon
from tf2mon.gameevent import GameEvent
from tf2mon.user import UserKey


class GameConnectedEvent(GameEvent):

    pattern = "(?P<username>.*) connected$"

    def handler(self, match: Match[str]) -> None:

        (username,) = match.groups()

        logger.log("CONNECT", tf2mon.users[UserKey(username)])

        tf2mon.ui.notify_operator = True
        if username == tf2mon.config.get("player_name"):
            tf2mon.reset_game()



"""Racist."""

import re
from pathlib import Path
from typing import Pattern

from loguru import logger

_RE_RACIST: Pattern[str] | None = None


def load_racist_data(path: Path) -> None:
    """Docstring."""

    global _RE_RACIST  # pylint: disable=global-statement

    logger.info(f"Reading `{path}`")
    lines = path.read_text(encoding="utf-8").splitlines()
    _RE_RACIST = re.compile("|".join(lines), flags=re.IGNORECASE) if len(lines) > 0 else None


def is_racist_text(text: str) -> bool:
    """Return True if this user appears to be racist."""

    return _RE_RACIST is not None and _RE_RACIST.search(text) is not None


def clean_username(username: str) -> str:
    """Return `username` cleaned of racism."""

    if _RE_RACIST is not None:
        m = _RE_RACIST.search(username)
        if m is not None:
            return m.string[: m.start()] + str("n" * (m.end() - m.start())) + m.string[m.end() :]
    return username



"""Scoreboard."""

import curses

import libcurses

import tf2mon
from tf2mon.player import Player
from tf2mon.texttable import TextColumn, TextTable
from tf2mon.user import Team, User


class Scoreboard:
    """Scoreboard."""

    table = TextTable(
        [
            TextColumn(-5, "UID"),
            TextColumn(1, "P"),
            TextColumn(2, "CC"),
            TextColumn(2, "SC"),
            # TextColumn(4, "CI"),
            TextColumn(-4, "AGE"),
            TextColumn(-10, "STEAMID"),
            TextColumn(-8, "CONN"),
            TextColumn(1, "S"),
            TextColumn(-3, "SNI"),
            TextColumn(-3, "K"),
            TextColumn(-3, "D"),
            TextColumn(4.1, "KD"),
            TextColumn(8, "CLASS"),
            TextColumn(25, "USERNAME"),
        ]
    )

    def __init__(
        self,
        win1: curses.window,
        color1: int,
        win2: curses.window,
        color2: int,
    ) -> None:
        """Create scoreboard for two teams.

        Args:
            win1:       curses window for team 1
            color1:     base color for team 2
            win2:       curses window for team 2
            color2:     base color for team 2
        """

        self.win1 = win1
        self.color1 = color1
        self.win2 = win2
        self.color2 = color2

        # calculate each column's x-ordinate for lookup by `heading`.
        self._col_x_width_by_heading = {}
        i = 0
        for col in self.table.columns:
            self._col_x_width_by_heading[col.heading] = (i, col.width)
            i += int(col.width) + 1

        # updated by `set_sort_order`
        self._sort_col_x = 0
        self._sort_col_width = 0

    def set_sort_order(self, sort_order: str) -> None:
        """Set sort order."""

        self._sort_col_x, self._sort_col_width = [
            int(x) for x in self._col_x_width_by_heading[sort_order]
        ]

    def refresh(self) -> None:
        """Display the scoreboards."""

        libcurses.clear_mouse_handlers()

        users = list(tf2mon.users.sorted())
        team1 = [x for x in users if x.team == Team.BLU]
        team2 = [x for x in users if x.team == Team.RED]

        # Fill in whatever space is left on the scoreboards with users
        # whose team is unknown; doesn't matter which side they're
        # displayed on.
        unassigned = [x for x in users if not x.team]
        nusers = self.win1.getmaxyx()[0] - 1
        while len(team1) < nusers and len(unassigned) > 0:
            team1.append(unassigned.pop(0))

        self._refresh_team(self.win1, self.color1, team1)
        self._refresh_team(self.win2, self.color2, team2 + unassigned)

    def _refresh_team(
        self,
        win: curses.window,
        color: int,
        team: list[User],
    ) -> None:

        # pylint: disable=too-many-branches
        # pylint: disable=too-many-locals

        ncols = win.getmaxyx()[1]
        win.erase()

        # column headings
        win.addnstr(0, 0, self.table.formatted_header, ncols, color)

        # highlight heading of active sort column
        win.chgat(
            0, self._sort_col_x, self._sort_col_width, color | curses.A_BOLD | curses.A_ITALIC
        )

        for lineno, user in enumerate(team, start=1):
            if not user.username:
                continue

            if user.dirty:
                user.dirty = False

                names = [user.username]
                _sp = user.steamplayer
                if user.perk:
                    names.append(user.perk)
                elif _sp:
                    if _sp.personaname and _sp.personaname != user.username:
                        names.append(_sp.personaname)
                    if _sp.realname:
                        names.append(_sp.realname)

                _steam_id = 0
                if user.steamid:
                    _steam_id = user.steamid.id

                _personastate = ""
                if _sp and _sp.personastate:
                    _personastate = str(_sp.personastate)

                _loccountrycode = ""
                if _sp and _sp.loccountrycode:
                    _loccountrycode = _sp.loccountrycode

                _locstatecode = ""
                if _sp and _sp.locstatecode:
                    _locstatecode = _sp.locstatecode

                _age = ""
                if _sp and _sp.age:
                    _age = str(_sp.age)

                user.last_scoreboard_line = self.table.format_detail(
                    user.userid,
                    _personastate,
                    _loccountrycode,
                    _locstatecode,
                    _age,
                    _steam_id,
                    user.s_elapsed,
                    user.n_status_checks,
                    user.nsnipes,
                    user.nkills,
                    user.ndeaths,
                    user.kdratio,
                    user.role.name,
                    " +".join(names),
                )

            try:
                win.addnstr(
                    lineno,
                    0,
                    user.last_scoreboard_line,
                    ncols,
                    tf2mon.ui.user_color(user, color),
                )
            except curses.error:
                break

            # register to handle mouse events on the user.
            begin_y, begin_x = win.getbegyx()
            libcurses.add_mouse_handler(
                self._onmouse, begin_y + lineno, begin_x + 0, ncols, user
            )

        win.noutrefresh()

    def _onmouse(self, mouse: libcurses.MouseEvent, user: User) -> bool:
        """Handle mouse events within team scoreboards."""

        if mouse.button != 1:
            return False  # not handled

        for active_user in tf2mon.users.active_users():
            active_user.selected = False
        user.selected = True
        tf2mon.ui.update_display()

        if mouse.nclicks == 2:
            user.kick(Player.CHEATER)
        elif mouse.nclicks == 3:
            user.kick(Player.RACIST)

        return True  # handled



"""Team Fortress 2 Console Monitor."""

import curses
import re
import threading
from pathlib import Path

import libcurses
from loguru import logger

import tf2mon
import tf2mon.game
from tf2mon.conlog import Conlog
from tf2mon.database import Database
from tf2mon.pkg import APPNAME
from tf2mon.player import Player
from tf2mon.racist import load_racist_data
from tf2mon.role import load_weapons_data
from tf2mon.steamplayer import SteamPlayer
from tf2mon.ui import UI


class Monitor:
    """Team Fortress 2 Console Monitor."""

    def run(self) -> None:
        """Run the Monitor."""
        libcurses.wrapper(self._run)

    def _run(self, win: curses.window) -> None:
        """Complete initialization; post CLI, options now available."""

        tf2mon.conlog = Conlog(tf2mon.options)
        load_weapons_data(Path(__file__).parent / "data" / "weapons.csv")
        load_racist_data(Path(__file__).parent / "data" / "racist.txt")
        tf2mon.ui = UI(win)
        tf2mon.controller.start()
        tf2mon.reset_game()

        # no need for threads if exiting at end of conlog
        if not tf2mon.options.follow:
            self.game()
            return

        # Read from conlog, write to display.
        thread = threading.Thread(name="GAME", target=self.game, daemon=True)
        thread.start()

        # Read from keyboard/mouse, write to display.
        self.admin()

    def game(self) -> None:
        """Read console log file and play game."""

        Database(tf2mon.options.database, [Player, SteamPlayer])
        assert tf2mon.conlog
        tf2mon.conlog.open()  # waits until it exists; then opens and returns.
        stepper = tf2mon.SingleStepControl

        while (line := tf2mon.conlog.readline()) is not None:
            # conlog.readline does not return excluded lines.
            if not line:
                continue

            event, match = None, None
            for event in [
                x
                for x in tf2mon.game.events + tf2mon.controller.controls
                if hasattr(x, "search") and x.search
            ]:
                assert event.search
                if match := event.search(line):
                    break
            else:
                logger.log("ignore", tf2mon.conlog.last_line)
                continue

            logger.log("regex", match)

            if hasattr(event, "start_stepping") and event.start_stepping:
                logger.log("ADMIN", f"break on {event.__class__.__name__}")
                stepper.start_single_stepping()

            elif stepper.pattern and stepper.pattern.search(line):
                pattern = stepper.pattern.pattern
                flags = "i" if (stepper.pattern.flags & re.IGNORECASE) else ""
                logger.log("ADMIN", f"break search /{pattern}/{flags}")
                stepper.start_single_stepping()

            level = "nextline" if stepper.is_stepping else "logline"
            logger.log(level, "-" * 80)
            logger.log(level, tf2mon.conlog.last_line)

            # check gate
            assert stepper.wait
            stepper.wait(None)
            if stepper.is_stepping:
                assert stepper.clear
                stepper.clear()

            if hasattr(event, "handler"):
                event.handler(match)
                tf2mon.MsgQueuesControl.send()
                tf2mon.ui.update_display()

    def admin(self) -> None:
        """Admin console read-evaluate-process-loop."""

        # pylint: disable=too-many-branches

        stepper = tf2mon.SingleStepControl
        assert stepper
        assert tf2mon.conlog

        while not tf2mon.conlog.is_eof or tf2mon.options.follow:

            tf2mon.ui.update_display()

            prompt = APPNAME
            if stepper.is_stepping:
                prompt += " (single-stepping)"
            prompt += ": "

            if (line := tf2mon.ui.getline(prompt)) is None:
                logger.log("console", "quit eof")
                return

            if line == "":  # enter
                if tf2mon.conlog.is_eof:
                    logger.log("console", f"lineno={tf2mon.conlog.lineno} <EOF>")
                # else:
                #     logger.trace("step...")
                assert stepper.set
                stepper.set()
                continue

            logger.log("console", f"line={line!r}")

            if "quit".find(line) == 0:
                logger.log("console", "quit")
                return

            try:
                cmd, arg = line.split(maxsplit=1)
            except ValueError:
                cmd = line
                arg = None

            if "breakpoint".find(cmd) == 0 and arg and arg.isdigit():
                stepper.set_single_step_lineno(int(arg))

            elif cmd[0] == "/":
                pattern = cmd[1:]
                if arg:
                    pattern += f" {arg}"
                stepper.set_single_step_pattern(pattern)

            elif "continue".find(cmd) == 0 or "go".find(cmd) == 0 or "run".find(cmd) == 0:
                stepper.stop_single_stepping()

            elif "kick".find(cmd) == 0 and arg and arg.isdigit():
                tf2mon.users.kick_userid(int(arg), Player.CHEATER)

            elif "kkk".find(cmd) == 0 and arg and arg.isdigit():
                tf2mon.users.kick_userid(int(arg), Player.RACIST)

            elif "suspect".find(cmd) == 0 and arg and arg.isdigit():
                tf2mon.users.kick_userid(int(arg), Player.SUSPECT)

            elif "dump".find(cmd) == 0:
                tf2mon.dump()

            elif "help".find(cmd) == 0:
                tf2mon.HelpControl.handler(None)

            elif "motd".find(cmd) == 0:
                tf2mon.MotdControl.handler(None)

            elif "pdb".find(cmd) == 0:
                tf2mon.debugger()

            else:
                logger.error(f"bad admin command {cmd!r}")



"""Cycle Grid Layout."""

from argparse import ArgumentParser
from enum import Enum
from typing import Match

import tf2mon
from tf2mon.control import CycleControl
from tf2mon.cycle import Cycle
from tf2mon.layouts.chat import ChatLayout
from tf2mon.layouts.default import DefaultLayout
from tf2mon.layouts.full import FullLayout
from tf2mon.layouts.tall import TallLayout
from tf2mon.layouts.tallchat import TallChatLayout
from tf2mon.layouts.wide import WideLayout


class GridLayoutControl(CycleControl):
    """Cycle grid layout."""

    name = "TOGGLE-LAYOUT"
    enum = Enum("enum", "CHAT DFLT FULL TALL MRGD WIDE")
    cycle = Cycle("_t_layout", enum)
    items = {
        enum.CHAT: ChatLayout,
        enum.DFLT: DefaultLayout,
        enum.FULL: FullLayout,
        enum.TALL: TallLayout,
        enum.MRGD: TallChatLayout,
        enum.WIDE: WideLayout,
    }

    def start(self) -> None:
        self.cycle.start(self.enum.__dict__[tf2mon.options.layout])
        tf2mon.ui.grid.handle_term_resized_event()

    def handler(self, _match: Match[str] | None) -> None:
        _ = self.cycle.next
        tf2mon.ui.grid.handle_term_resized_event()
        tf2mon.ui.update_display()

    def add_arguments_to(self, parser: ArgumentParser) -> None:
        arg = parser.add_argument(
            "--layout",
            choices=[x.name for x in list(self.enum)],
            default="MRGD",
            help="choose display layout",
        )
        self.add_fkey_to_help(arg)
        self.cli.add_default_to_help(arg)



"""Chats controls."""

from typing import Match

from loguru import logger

import tf2mon
from tf2mon.chat import Chat
from tf2mon.control import Control


class ChatsControl(Control):
    """Chats control."""

    _chats: list[Chat] = []

    def append(self, chat: Chat) -> None:
        self._chats.append(chat)
        tf2mon.ui.show_chat(chat)

    def clear(self) -> None:
        self._chats = []
        self.refresh()

    def refresh(self) -> None:

        if not hasattr(tf2mon.ui, "layout"):
            return
        assert tf2mon.ui.layout

        if tf2mon.ui.layout.chatwin_blu:
            tf2mon.ui.layout.chatwin_blu.erase()
        if tf2mon.ui.layout.chatwin_red:
            tf2mon.ui.layout.chatwin_red.erase()
        for chat in self._chats:
            tf2mon.ui.show_chat(chat)
        if tf2mon.ui.layout.chatwin_blu:
            tf2mon.ui.layout.chatwin_blu.noutrefresh()
        if tf2mon.ui.layout.chatwin_red:
            tf2mon.ui.layout.chatwin_red.noutrefresh()


class ClearChatsControl(Control):
    """Clear chat window(s)."""

    name = "CLEAR-CHATS"

    def handler(self, _match: Match[str] | None = None) -> None:
        tf2mon.ChatsControl.clear()
        tf2mon.ui.update_display()
        logger.success(self.name)


class RefreshChatsControl(Control):
    """Refresh chat window(s)."""

    name = "REFRESH-CHATS"

    def handler(self, _match: Match[str] | None = None) -> None:
        tf2mon.ChatsControl.refresh()
        tf2mon.ui.update_display()
        logger.success(self.name)



# mypy: ignore-errors
# This file is no longer used.

from pathlib import Path

import pytest

from tf2mon.database import Database
from tf2mon.player import Player
from tf2mon.steamplayer import SteamPlayer


@pytest.fixture(scope="session")
def session():

    return Database(Path(".cache/tf2mon.db"), [Player, SteamPlayer])



"""Table/Column formatters."""

from dataclasses import dataclass, field
from typing import Any


@dataclass
class TextColumn:
    """Column formatter."""

    width: int | float
    heading: str
    fmt_heading: str = field(default_factory=str, init=False)
    fmt_detail: str = field(default_factory=str, init=False)

    def __post_init__(self) -> None:
        """Init table column."""

        if isinstance(self.width, float):
            _width, _precision = [int(x) for x in str(self.width).split(".")]
            self.width = _width
            _right = True
        elif isinstance(self.width, int):
            _precision = None
            if self.width < 0:
                _width = self.width = abs(self.width)
                _right = True
            else:
                _width = self.width
                _right = False
        else:
            raise TypeError("width must be int|float")

        # build format string for the header
        self.fmt_heading = "{:"
        if _right:
            self.fmt_heading += ">"
        self.fmt_heading += str(_width) + "}"

        # build format string for the detail
        self.fmt_detail = "{:"
        if _right:
            self.fmt_detail += ">"
        self.fmt_detail += str(_width)
        if _precision:
            self.fmt_detail += f".{_precision}f"
        self.fmt_detail += "}"


@dataclass
class TextTable:
    """Table formatter."""

    columns: list[TextColumn] = field(default_factory=list)
    _formatted_header: str = field(default_factory=str, init=False)
    _fmt_detail: str = field(default_factory=str, init=False)

    @property
    def formatted_header(self) -> str:
        """Docstring."""

        if not self._formatted_header:
            fmt = " ".join([x.fmt_heading for x in self.columns])
            self._formatted_header = fmt.format(*[x.heading for x in self.columns])

        return self._formatted_header

    def format_detail(self, *values: Any) -> str:
        """Docstring."""

        result = []
        for column, value in zip(self.columns, values):
            if value is None:
                result.append(column.fmt_heading.format(""))
            else:
                result.append(column.fmt_detail.format(value))

        return " ".join(result)



import logging
from pathlib import Path

import pytest
import tomli

from tf2mon.steamplayer import SteamPlayer
from tf2mon.steamweb import SteamWebAPI

logging.basicConfig(force=True, level=logging.DEBUG)


@pytest.fixture(name="api", scope="session")
def api_(session: str) -> SteamWebAPI:
    # pylint: disable=unused-argument

    path = Path("~/.tf2mon.toml").expanduser()
    config = tomli.loads(path.read_text(encoding="utf-8"))
    webapi_key = None
    tf2mon = config.get("tf2mon")
    assert tf2mon
    webapi_key = tf2mon.get("webapi_key")
    return SteamWebAPI(webapi_key)


@pytest.mark.parametrize(("steamid"), [-3, -2, -1, 0])
def test_fetch_steamid_not_found(api: SteamWebAPI, steamid: int) -> None:
    result = api.fetch_steamid(steamid)
    # print(result)
    assert result
    assert result.personaname == "???"


@pytest.mark.parametrize(
    ("steamid", "expected"),
    [
        (
            2,
            SteamPlayer(
                steamid=2,
                personaname="alfred",
                profileurl="https://steamcommunity.com/id/zoe/",
                personastate=0,
                realname="Alfred",
                timecreated=1063193241,
                loccountrycode="",
                locstatecode="",
                loccityid="",
            ),
        ),
    ],
)
def test_fetch_known_steamids(api: SteamWebAPI, steamid: int, expected: SteamPlayer) -> None:
    result = api.fetch_steamid(steamid)
    assert result
    assert result.steamid == expected.steamid
    assert result.personaname == expected.personaname
    assert result.profileurl == expected.profileurl
    assert result.personastate == expected.personastate
    assert result.realname == expected.realname
    assert result.timecreated == expected.timecreated
    assert result.loccountrycode == expected.loccountrycode
    assert result.locstatecode == expected.locstatecode
    assert result.loccityid == expected.loccityid
    # assert result.mtime == expected.mtime
    # print(f"result={result}")


@pytest.mark.parametrize(("steamid"), [42708103])
def test_fetch_steamid_found(api: SteamWebAPI, steamid: int) -> None:
    steamplayer = api.fetch_steamid(steamid)
    assert steamplayer
    # print(steamplayer)



"""Function Keys."""

from __future__ import annotations

import curses
import curses.ascii
import re
from dataclasses import dataclass, field
from typing import Any, ClassVar, Pattern

# Values for `modifier`; also used as `shortname` prefix.
MOD_BASE = ""
MOD_CTRL = "c"
MOD_SHIFT = "s"


class FKey:
    """The simultaneous pressing of a base key and an optional modifier key."""

    pattern: ClassVar[Pattern[str]] = re.compile(r"^(?:(SHIFT|CTRL)\+)?(F)?(.+)$")

    curses_from_game: ClassVar[dict[str, int]] = {
        "KP_HOME": curses.KEY_HOME,
        "KP_END": curses.KEY_END,
        "KP_UPARROW": curses.KEY_UP,
        "KP_DOWNARROW": curses.KEY_DOWN,
        "KP_LEFTARROW": curses.KEY_LEFT,
        "KP_RIGHTARROW": curses.KEY_RIGHT,
        "KP_PGUP": curses.KEY_PPAGE,
        "KP_PGDN": curses.KEY_NPAGE,
        "KP_5": curses.KEY_B2,
        "KP_INS": curses.KEY_IC,
        "KP_DEL": curses.KEY_DC,
        # "xxx": curses.KEY_SR,
        # "xxx": curses.KEY_SF,
        # "xxx": curses.KEY_SLEFT,
        # "xxx": curses.KEY_SRIGHT,
        # "xxx": curses.KEY_SDC,  # can't do KEY_SIC
    }

    # physical keys, by `name`.
    pkeys: ClassVar[dict[str, PKey]] = {}

    # Keyname in TF2 terms, with optional modifier "shift+" or "ctrl+" (not both).
    # e.g., "A", "shift+A", "ctrl+KP_LEFTARROW".
    keyspec: str

    name: str  # TF2 terms: "B", "KP_LEFTARROW", "F1"
    modifier: str  # MOD_BASE, MOD_CTRL or MOD_SHIFT
    key: int  # ord("B"), curses.KEY_LEFT, curses.KEY_F1
    shortname: str  # for status-line.
    longname: str  # for `--help`.
    pkey: PKey  # physical key

    def __init__(self, keyspec: str):
        """Init `FKey` from `keyspec`.

        Keyname in TF2 terms, with optional modifier "shift+" or "ctrl+" (not both).
        e.g., "A", "shift+A", "ctrl+KP_LEFTARROW".

        Case-insensitive:
        >>> FKey("a").__dict__ == FKey("A").__dict__
        True
        >>> FKey("shift+a").__dict__ == FKey("shift+A").__dict__
        True

        Note that:
        >>> FKey("shift+a").__dict__ != FKey("A").__dict__
        True
        >>> FKey("shift+A").__dict__ != FKey("A").__dict__
        True
        """

        # pylint: disable=too-many-branches

        if not keyspec:
            raise ValueError("keyspec", keyspec)
        self.keyspec = keyspec.upper()

        matched = self.pattern.match(self.keyspec)
        if not matched:
            raise ValueError("keyspec", self.keyspec)

        mod, eff, key = matched.groups()
        self.name = (eff or "") + key  # TF2 terms: "B", "KP_LEFTARROW", "F1"

        self.modifier = MOD_BASE
        if mod:
            if mod == "SHIFT":
                self.modifier = MOD_SHIFT
            elif mod == "CTRL":
                self.modifier = MOD_CTRL
            else:
                raise ValueError("keyspec", self.keyspec)

        if eff:
            self.key = curses.KEY_F0
            try:
                self.key += int(key)
            except ValueError as exc:
                raise ValueError("keyspec", self.keyspec) from exc
        elif len(key) == 1:
            self.key = ord(key)
        elif (key := self.curses_from_game.get(key)) is not None:
            self.key = key
        else:
            raise ValueError("keyspec", self.keyspec)

        if eff:
            if self.modifier == MOD_SHIFT:
                self.key += 12
            elif self.modifier == MOD_CTRL:
                self.key += 24

        self.shortname = self.modifier + self.name  # display name

        self.longname = ""
        if self.modifier == MOD_SHIFT:
            self.longname = "shift+"
        elif self.modifier == MOD_CTRL:
            self.longname = "ctrl+"
        self.longname += self.name

        pkey = self.__class__.pkeys.get(self.name)
        if not pkey:
            pkey = PKey(self.name)
            self.__class__.pkeys[self.name] = pkey
        self.pkey = pkey

    def __repr__(self) -> str:
        return str(self.__dict__)

    @property
    def is_base(self) -> bool:
        """Return True if only `base` key was pressed."""
        return self.modifier == MOD_BASE

    @property
    def is_ctrl(self) -> bool:
        """Return True if `ctrl` was also pressed."""
        return self.modifier == MOD_CTRL

    @property
    def is_shift(self) -> bool:
        """Return True if `shift` was also pressed."""
        return self.modifier == MOD_SHIFT

    @property
    def is_ascii(self) -> bool:
        """Return True if this is an ascii key."""
        return curses.ascii.isascii(self.key)

    def bind(self, payload: Any) -> None:
        """Bind `payload` to this key."""

        if self.is_ctrl:
            if self.pkey.ctrl:
                raise ValueError("duplicate keyspec", self.keyspec)
            self.pkey.ctrl = payload
        #
        elif self.is_shift:
            if self.pkey.shift:
                raise ValueError("duplicate keyspec", self.keyspec)
            self.pkey.shift = payload
        #
        else:
            if self.pkey.base:
                raise ValueError("duplicate keyspec", self.keyspec)
            self.pkey.base = payload


@dataclass
class PKey:
    """A physical key may perform `base`, `ctrl` and `shift` `Function`s."""

    name: str  # e.g., "A", "F1"
    base: Any | None = field(default=None, init=False)
    ctrl: Any | None = field(default=None, init=False)
    shift: Any | None = field(default=None, init=False)

    @property
    def bindings(self) -> list[Any]:
        """Return list of `payload`s bound to this `PKey`."""
        return [x for x in [self.base, self.ctrl, self.shift] if x is not None]



from typing import Match

from loguru import logger

import tf2mon
from tf2mon.gameevent import GameEvent
from tf2mon.user import UserKey


class GameSuicideEvent(GameEvent):

    pattern = "(?P<username>.*) suicided.$"

    def handler(self, match: Match[str]) -> None:

        (username,) = match.groups()

        user = tf2mon.users[UserKey(username)]
        logger.log("SUICIDE", user)
        user.ndeaths += 1



"""Application control."""

from __future__ import annotations

import argparse
import re
from typing import Any, ClassVar, Match

import tf2mon
from tf2mon.cli import Tf2monCLI
from tf2mon.cycle import Cycle
from tf2mon.fkey import FKey
from tf2mon.pkg import APPTAG


class Control:
    """Application control.

    Optional methods that derived classes MAY define:
    :g/[gh]..attr/p

    CONTROL -
        status
            Return current value as `str` formatted for display.
            Default `None` to hide.

        handler
            Perform monitor function.

        name
            Command name; e.g., 'HELP', 'TOGGLE-SORT'.

        action
            Game "script" text to `exec`.

        add_arguments_to
            Add control to cli `parser`.

        start
            Finalize initialization after curses has been started.
    """

    pkeys: dict[str, FKey] = {}

    name: str = ""  # token
    # status: Callable[..., str] = None
    # handler: Callable[[re.Match], None] = None
    action: str = ""

    # Optional; function key to operate control.
    fkey: FKey | None = None

    #
    cli: ClassVar[Tf2monCLI]

    #
    match = None
    search = None

    def __init__(self) -> None:
        """Init."""

        if self.name:
            # Default action, have game send this event notification
            # message to monitor whenever game calls for this command,
            # such as in response to an in-game key-press or mouse-click.
            token = APPTAG + self.name
            if not self.action:
                self.action = f"echo {token}"
            pattern = f"^{token}$"
            self._re = re.compile(pattern)
            self.match = self._re.match
            self.search = self._re.search

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__dict__})"

    def add_fkey_to_help(self, arg: argparse.Action) -> None:
        """Add fkey to help text for `arg`."""

        if not self.fkey:
            return
        text = f" (fkey: `{self.fkey.longname}`)"
        assert arg.help
        assert self.cli
        if arg.help.endswith(self.cli.help_line_ending):
            arg.help = (
                arg.help[: -len(self.cli.help_line_ending)] + text + self.cli.help_line_ending
            )
        else:
            arg.help += text

    def toggling_enabled(self) -> bool:
        """Return True if toggling is enabled.

        Don't allow toggling when replaying a game (`--rewind`),
        unless `--allow-toggles` is also given... or if single-stepping

        This is checked by keys that alter the behavior of gameplay;
        it is not checked by keys that only alter the display.
        """

        assert tf2mon.conlog

        return (
            tf2mon.conlog.is_eof
            or tf2mon.options.allow_toggles
            or tf2mon.SingleStepControl.is_stepping
        )


class BoolControl(Control):
    """Bool control."""

    cycle: Cycle

    def handler(self, _match: Match[str] | None) -> None:
        """Handle event."""

        if self.toggling_enabled():
            _ = self.cycle.next
            tf2mon.ui.show_status()

    def status(self) -> str:
        """Return value formatted for display."""

        return self.cycle.name.upper() if self.cycle.value else self.cycle.name

    @property
    def value(self) -> bool:
        """Return value."""

        return bool(self.cycle.value)


class CycleControl(Control):
    """Cycle control."""

    cycle: Cycle
    items: dict[Any, Any] = {}

    def status(self) -> str:
        """Return value formatted for display."""

        return str(self.cycle.value.name)

    @property
    def value(self) -> Any:
        """Return value."""

        return self.items[self.cycle.value]



"""Database."""

from __future__ import annotations

import dataclasses
import sqlite3
from pathlib import Path
from typing import Any, Iterator

from loguru import logger

DATABASE: sqlite3.dbapi2.Cursor | None = None


def Database(  # pylint: disable=invalid-name
    path: Path | None = None,
    tables: list[type[DatabaseTable]] | None = None,
) -> sqlite3.dbapi2.Cursor | None:  # noqa invalid-name
    """Open and return new session with database."""

    global DATABASE  # pylint: disable=global-statement
    if not DATABASE and path:

        _path = path.expanduser()
        logger.info(f"Opening `{_path}`")
        conn = sqlite3.connect(_path, check_same_thread=False)
        conn.row_factory = sqlite3.Row
        DATABASE = conn.cursor()

        for table in tables or []:
            table.create_table()

    return DATABASE


class DatabaseTable:
    """Base class for all database tables."""

    __tablename__: str

    @classmethod
    def create_table(cls) -> None:
        """Docstring."""

        raise NotImplementedError

    @classmethod
    def select_all(cls) -> Iterator[object]:
        """Yield all rows in table."""

        db = Database()
        assert db

        for row in db.execute(f"select * from {cls.__tablename__}"):
            yield cls(*tuple(row))

    @classmethod
    def valueholders(cls) -> str:
        """Return text for sql `values` clause."""

        placeholders = ",".join(["?"] * len(dataclasses.fields(cls)))  # type: ignore
        return f"values({placeholders})"

    def astuple(self) -> tuple[Any, ...]:
        """Return column values."""

        return dataclasses.astuple(self)  # type: ignore

    def upsert(self) -> None:
        """Update or Insert this row into the table."""

        db = Database()
        assert db

        try:
            db.execute(
                f"replace into {self.__tablename__} {self.valueholders()}",
                self.astuple(),
            )
        except Exception as err:
            logger.critical(err)
            raise
        db.connection.commit()



from typing import Match

from loguru import logger

import tf2mon
from tf2mon.gameevent import GameEvent
from tf2mon.user import Team


class GameServerEvent(GameEvent):

    # must be before `chat`
    # account : not logged in  (No account specified)
    # version : 6173888/24 6173888 secure
    # map     : pl_barnblitz at: 0 x, 0 y, 0 z
    # udp/ip  : 208.78.164.167:27067  (public ip: 208.78.164.167)
    # tags    : hidden,increased_maxplayers,payload,valve
    # steamid : [A:1:3814649857:15826] (90139968514486273)
    # players : 20 humans, 0 bots (32 max)
    # edicts  : 1378 used of 2048 max

    pattern = r"(account|version|map|udp\/ip|tags|steamid|players|edicts)\s+: (.*)"

    def handler(self, _match: Match[str] | None) -> None:
        pass  # logger.log("server", m.group(0)),


class GamePingEvent(GameEvent):

    # "06/05/2022 - 13:54:19:   67 ms : luft"
    # "06/05/2022 - 13:54:19:xy 87 ms : BananaHatTaco"
    pattern = r"\s*\d+ ms .*"

    def handler(self, _match: Match[str] | None) -> None:
        pass  # logger.log("server", m.group(0)),


class GameLobbyFailedEvent(GameEvent):

    pattern = "Failed to find lobby shared object"

    def handler(self, match: Match[str]) -> None:
        logger.trace("tf_lobby_debug failed: " + match.group(0))


class GameTeamsSwitchedEvent(GameEvent):

    pattern = "^Teams have been switched"

    def handler(self, _match: Match[str] | None) -> None:
        pass  # tf2mon.users.switch_teams()


class GameUserSwitchedEvent(GameEvent):

    pattern = r"You have switched to team (?P<teamname>\w+) and will"

    def handler(self, match: Match[str]) -> None:
        tf2mon.users.my.team = Team(match.group("teamname"))


class GameHostnameEvent(GameEvent):

    # hostname: Valve Matchmaking Server (Virginia iad-1/srcds148 #53)

    pattern = "^hostname: (.*)"

    def handler(self, _match: Match[str] | None) -> None:
        tf2mon.users.check_status()



"""Miscelleneous controls."""

import textwrap
from pathlib import Path
from typing import Match

import tf2mon
from tf2mon.control import Control


class HelpControl(Control):
    """Display Help."""

    name = "HELP"

    def handler(self, _match: Match[str] | None) -> None:

        tf2mon.ui.show_journal("help", " Function Keys ".center(80, "-"))
        for line in tf2mon.controller.fkey_help().splitlines():
            tf2mon.ui.show_journal("help", line)

        tf2mon.ui.show_journal("help", " Admin Commands ".center(80, "-"))
        for line in (
            textwrap.dedent(
                """
            Press Enter to process next line.
            Enter "b 500" to set breakpoint at line 500.
            Enter "/pattern[/i]" to set search pattern.
            Enter "/" to clear search pattern.
            Enter "c" to continue.
            Enter "quit" or press ^D to quit."
                """
            )
            .strip()
            .splitlines()
        ):
            tf2mon.ui.show_journal("help", line)

    def status(self) -> str:
        return self.name


class MotdControl(Control):
    """Display Message of the Day."""

    name = "MOTD"
    _path: Path

    def start(self) -> None:
        self._path = tf2mon.options.tf2_install_dir / "cfg" / "motd.txt"

    def handler(self, _match: Match[str] | None) -> None:
        tf2mon.ui.show_journal("help", f" {self._path} ".center(80, "-"))

        with open(self._path, encoding="utf-8") as file:
            for line in file:
                tf2mon.ui.show_journal("help", line.strip())



from typing import Match

from loguru import logger

import tf2mon
from tf2mon.gameevent import GameEvent
from tf2mon.user import Team, UserKey


class GameCaptureEvent(GameEvent):

    pattern = r"(?P<username>.*) (?P<action>(?:captured|defended)) (?P<capture_pt>.*) for team #(?P<s_teamno>\d)$"

    def handler(self, match: Match[str]) -> None:

        username, action, capture_pt, s_teamno = match.groups()

        for name in username.split(", "):  # fix: names containing commas

            user = tf2mon.users[UserKey(name)]

            try:
                team = Team(int(s_teamno))
                if user.team and user.team != team:
                    logger.warning(f"{user} switched to team `{team.name}`")
                user.team = team
            except ValueError as err:
                logger.error(f"{err} s_teamno {s_teamno!r}")

            if action == "captured":
                user.ncaptures += 1
                level = "CAP"
            else:
                user.ndefenses += 1
                level = "DEF"

            user.dirty = True

            assert user.team
            level += user.team.name

            logger.log(level, f"{user} {capture_pt!r}")
            user.actions.append(f"{level} {capture_pt!r}")



"""Message queues control."""

from pathlib import Path
from typing import IO, Match

from loguru import logger

import tf2mon
from tf2mon.control import Control


class MsgQueuesControl(Control):
    """Message queues control."""

    _controls: list[Control] = []
    _file: IO[str] | None = None

    def start(self) -> None:
        """Complete initialization; post CLI, options now available."""

        self._controls.append(tf2mon.KicksControl)
        self._controls.append(tf2mon.SpamsControl)

        # Location of TF2 `exec` scripts.
        _scripts = tf2mon.options.tf2_install_dir / "cfg" / "user"

        # MsgQueue aliases; written often; left open.
        _dynamic_path = _scripts / "tf2mon-pull.cfg"

        # Static aliases and key bindings; written once (now).
        _static_path = _scripts / "tf2mon.cfg"

        if not _scripts.is_dir():
            logger.warning(f"Can't find scripts dir `{_scripts}`")
            logger.warning(f"Not opening `{_dynamic_path}`")
            logger.warning(f"Not writing `{_static_path}`")
            return

        logger.info(f"Writing `{_static_path}`")
        script = tf2mon.controller.as_tf2_exec_script(
            str(_static_path.relative_to(_scripts.parent)),
            str(_dynamic_path.relative_to(_scripts.parent)),
        )
        _static_path.write_text(script, encoding="utf-8")

        logger.info(f"Opening `{_dynamic_path}`")
        # pylint: disable=consider-using-with
        self._file = open(_dynamic_path, "w", encoding="utf-8")  # noqa

    def clear(self) -> None:
        """Clear all message queues."""

        for control in self._controls:
            assert hasattr(control, "clear")
            control.clear()

    def send(self) -> None:
        """Send data to tf2 by writing aliases to an `exec` script."""

        if not self._file:
            return

        self._file.seek(0)
        self._file.truncate()
        for control in self._controls:
            assert hasattr(control, "aliases")
            print("\n".join(control.aliases()), file=self._file)
        self._file.flush()


class DisplayFileControl(Control):
    """Display file."""

    name = "DISPLAY-FILE"
    _path: Path | None = None

    def start(self) -> None:
        self._path = tf2mon.options.tf2_install_dir / "cfg" / "user" / "tf2mon.cfg"

    def handler(self, _match: Match[str] | None) -> None:

        assert self._path
        tf2mon.ui.popup(
            "help",
            f" {self._path} ".center(80, "-") + "\n" + self._path.read_text(encoding="utf-8"),
        )



"""Command line interface."""

import threading
from pathlib import Path

import xdg
from libcli import BaseCLI
from loguru import logger

import tf2mon
import tf2mon.layouts
from tf2mon._logger import configure_logger
from tf2mon.conlog import Conlog
from tf2mon.database import Database
from tf2mon.monitor import Monitor
from tf2mon.steamweb import SteamWebAPI

__all__ = ["Tf2monCLI"]


class Tf2monCLI(BaseCLI):
    """Command line interface."""

    _cachedir = BaseCLI.hideuser(xdg.xdg_cache_home() / __package__)

    config = {
        # name of config file.
        "config-file": Path("~/.tf2mon.toml"),
        # toml [section-name].
        "config-name": "tf2mon",
        #
        # location of game to monitor
        "tf2_install_dir": Path("~/SteamLibrary/steamapps/common/Team Fortress 2/tf"),
        #
        # basename relative to `tf2_install_dir`.
        "con_logfile": Path("console.log"),
        #
        # databases.
        "database": _cachedir / "tf2mon.db",
        "hackers": _cachedir / "hackers.json",
        "exclude-file": BaseCLI.hideuser(Path(__file__).parent / "data" / "exclude.txt"),
        "webapi_key": "",
        # this player.
        "player_name": "Bad Dad",
    }

    def __init__(self, argv: list[str] | None = None) -> None:
        """Build and parse command line."""

        threading.current_thread().name = "MAIN"
        self.monitor = Monitor()
        super().__init__(argv)

    def init_logging(self, verbose: int) -> None:
        """Set logging levels based on `--verbose`."""

        init_logging_called = self.init_logging_called
        super().init_logging(verbose)
        if not init_logging_called:
            configure_logger()

    def init_parser(self) -> None:
        """Initialize argument parser."""

        self.parser = self.ArgumentParser(
            prog=__package__,
            description=self.dedent(
                """
    Team Fortress II (`TF2`) Console Monitor, `%(prog)s`, is an interactive
    terminal application that displays scoreboards and player statistics of
    an active game. `%(prog)s` can also `--rewind` and `--single-step`
    through previous games.

    `%(prog)s` recognizes and tracks abusive players, such as racists and
    `cathook-bots`. It provides application and in-game key-bindings to
    `CALLVOTE KICK` cheaters/hackers, or mark as racist or suspect.

    `%(prog)s` reads TF2's console logfile (`con_logfile`), creates action
    scripts in TF2's `cfg/user` directory, and binds keys for gamer to
    press to take those actions, such as issue `SAY` and `CALLVOTE KICK`
    commands.

    Other in-game key-bindings taunt gamer's last victim/killer with a
    `CHAT` message customized with their name, k/d ratio, rotating spam,
    weapon and duel-scores (`Taunt` and `Throe`).

    By default, `%(prog)s` starts reading `con_logfile` from its end
    (`--no-rewind`), and `--follow`s its tail.
                """
            ),
        )

    def add_arguments(self) -> None:
        """Add arguments to parser."""

        self._add_base_args()
        tf2mon.controller.add_arguments_to(self.parser)
        self._add_debug_args()
        self._add_database_args()
        self._add_fkeys_args()
        self._add_numpad()
        self._add_duels()
        self._add_operate()
        self._add_terminal()
        self._add_windows()
        self._add_scoreboard()
        self._add_logfiles()

    def _add_base_args(self) -> None:

        arg = self.parser.add_argument(
            "--tf2-install-dir",
            metavar="DIR",
            default=Path(self.config["tf2_install_dir"]),
            type=Path,
            help="TF2 installation directory",
        )
        self.add_default_to_help(arg)

        rewind = self.parser.add_mutually_exclusive_group()
        arg = rewind.add_argument(
            "--rewind",
            dest="rewind",
            action="store_true",
            default=False,
            help="start from head of logfile",
        )
        self.add_default_to_help(arg)

        arg = rewind.add_argument(
            "--no-rewind",
            dest="rewind",
            action="store_false",
            help="start from tail of logfile",
        )
        self.add_default_to_help(arg)

        follow = self.parser.add_mutually_exclusive_group()
        arg = follow.add_argument(
            "--follow",
            dest="follow",
            action="store_true",
            default=True,
            help="follow end of logfile forever",
        )
        self.add_default_to_help(arg)

        arg = follow.add_argument(
            "--no-follow",
            dest="follow",
            action="store_false",
            help="exit at end of logfile",
        )
        self.add_default_to_help(arg)

        arg = self.parser.add_argument(
            "con_logfile",
            default=Path(self.config["con_logfile"]),
            nargs="?",
            type=Path,
            help="TF2 console logfile; relative to `--tf2-install-dir`",
        )
        self.add_default_to_help(arg)

        self.parser.add_argument(
            "--list-con-logfile",
            action="store_true",
            help="show path to logfile and exit",
        )

        self.parser.add_argument(
            "--trunc-con-logfile",
            action="store_true",
            help="truncate logfile and exit",
        )

        self.parser.add_argument(
            "--clean-con-logfile",
            action="store_true",
            help="filter-out excluded lines from logfile to stdout and exit",
        )

        arg = self.parser.add_argument(
            "--exclude-file",
            metavar="FILE",
            default=Path(self.config["exclude-file"]),
            type=Path,
            help="exclude lines that match patterns in `FILE`",
        )
        self.add_default_to_help(arg)

    def _add_debug_args(self) -> None:

        group = self.parser.add_argument_group("Debugging options")

        group.add_argument(
            "--single-step",
            action="store_true",
            help="single-step at startup",
        )

        group.add_argument(
            "--break",
            dest="breakpoint",
            type=int,
            metavar="LINENO",
            help="single-step at line `LINENO`",
        )

        group.add_argument(
            "--search",
            metavar="PATTERN",
            help="single-step when line matches `PATTERN`; add `/i` to ignore case.",
        )

        group.add_argument(
            "--inject-cmd",
            dest="inject_cmds",
            metavar="LINENO:CMD",
            action="append",
            help="inject `CMD` before line `LINENO`",
        )

        group.add_argument(
            "--inject-file",
            metavar="FILE",
            help="read list of inject commands from `FILE`",
        )

        arg = group.add_argument(
            "--allow-toggles",
            action="store_true",
            help="allow toggles when `--rewind`",
        )
        self.add_default_to_help(arg)

    def _add_database_args(self) -> None:

        group = self.parser.add_argument_group("Database options")

        arg = group.add_argument(
            "--database",
            metavar="FILE",
            default=Path(self.config["database"]),
            type=Path,
            help="main database",
        )
        self.add_default_to_help(arg)

        arg = group.add_argument(
            "--hackers",
            metavar="FILE",
            default=Path(self.config["hackers"]),
            type=Path,
            help="hackers database",
        )
        self.add_default_to_help(arg)

        group.add_argument(
            "--print-steamids",
            nargs="+",
            metavar="STEAMID",
            help="print `ISteamUser.GetPlayerSummaries` for `STEAMID` and exit",
        )

        group.add_argument(
            "--print-hackers",
            action="store_true",
            help="print hackers database and exit",
        )

        self.parser.add_argument_group(
            "Configuration file",
            self.dedent(
                """
    The configuration file (see `--config FILE` below) defines local
    settings:

        [%(prog)s]
        tf2_install_dir = "/path/to/your/tf2/installation"
        webapi_key = "your-steamworks-webapi-key"
        player_name = "Your Name"
                """
            ),
        )

    def _add_numpad(self) -> None:

        self.parser.add_argument_group(
            "In-Game Controls, Numpad",
            self.dedent(
                """
    While playing TF2, use the `Numpad` to kick cheaters, and generate
    `Taunt` and death-`Throe` spam.

    Messages are placed into queues, and may be popped off either end.

        Queues -->    Kicks     Admin     Spams
                        |         |         |
                        v         v         v
                   +---------+---------+---------+
        last/      |         |         |         |
        newest --> |   pop   |         |   pop   |
                   |         |         |         |
                   +---------+---------+---------+
                   |         |         |         |
                   |  clear  |  clear  |  clear  |
                   |         |  both   |         |
                   +---------+---------+---------+
        first/     |         |         |         |
        oldest --> | popleft |  vet    | popleft |
                   |         |         |         |
                   +---------+---------+---------+

    To vet new players that have joined the game, `%(prog)s` needs their
    `steamid`s, and to get them, gamer must press `NUMPAD-DOWNARROW`. The
    `status` panel will be highlighted when the monitor needs steamids.

    When detected, `%(prog)s` pushes hackers onto the `Kicks` queue, and
    alerts the gamer, who may then press `HOME`/`END` to issue `CHAT`
    and `CALLVOTE KICK` commands.

    When gamer kills an opponent, `%(prog)s` pushes a `Taunt` onto the
    `Spams` queue; on death, a `Throe`. Enable/disable with `F3`. Send
    with `PGUP`/`PGDN`.

    The monitor can only push actions onto the queues; gamer must pop
    for action to be taken, or clear to discard.
                """
            ),
        )

    def _add_duels(self) -> None:

        self.parser.add_argument_group(
            "Duels",
            self.dedent(
                """
    The user-panel displays battles with opponents, grouped by weapon (and
    its state when fired).

        ┌─────────────────────────────────────────────────────────────────────┐
        │Duels:                                                               │
        │ 4 and  2 vs GLaDOS                                                  │
        │             K  2 soldier        quake_rl                            │
        │             K  1 pyro           flamethrower                        │
        │             K  1 pyro     +crit flamethrower         +Low Gravity   │
        │             D  2 heavy          minigun                             │
        │ 5 and  0 vs CreditToTeam                                            │
        │             K  2 soldier        quake_rl                            │
        │             K  2 pyro     +crit flamethrower                        │
        │             K  1 pyro           flamethrower                        │
        │ 3 and  1 vs Aperture Science Prototype XR7                          │
        │             K  1 soldier        world                               │
        │             K  1 soldier        quake_rl             +Invisibility  │
        │             K  1 pyro     +crit flamethrower                        │
        │             D  1 demo           player                              │
        └─────────────────────────────────────────────────────────────────────┘
                """
            ),
        )

    def _add_fkeys_args(self) -> None:

        self.parser.add_argument_group(
            "Function Keys",
            "These function keys are available in-game and in the monitor:\n\n"
            + tf2mon.controller.fkey_help(),
        )

    def _add_operate(self) -> None:

        self.parser.add_argument_group(
            "Where to Operate",
            self.dedent(
                """
    `%(prog)s` works by reading the `con_logfile` to which `TF2` logs
    messages during the game. `%(prog)s` can either "tail -f" an active
    game, or `--rewind` and replay saved logfiles. Press `Enter` in the
    admin console to process the next line when in `--single-step` mode.
    Type `quit` or press `^D` to exit.

        `One-machine, Two-monitors`
            Run `%(prog)s` on a secondary monitor, while playing game on
            primary monitor.

        `Two-machines, SSH`
            ssh from another machine and run.

        `Two-machines, NFS`
            cross-mount TF2's `cfg` tree to another box and run from there.
                """
            ),
        )

    def _add_terminal(self) -> None:

        self.parser.add_argument_group(
            "Terminal Size",
            self.dedent(
                """
    `%(prog)s` requires a large terminal. Maximize the window, and use keys
    (maybe `Ctrl-Minus` and `Shift-Ctrl-Plus`) to resize. The wider the
    terminal, the more player data will be displayed:

        36x146 minimum
        42x173 display personaname
        52x211 display realname
        62x272 display age/location
                """
            ),
        )

    def _add_windows(self) -> None:
        self.parser.add_argument_group(
            "Resizable Windows",
            self.dedent(
                """
    `Drag-and-drop` an interior border to resize the windows on either side.

    `Double-click` an interior border to enter `Resize Mode`.
        `scroll-wheel` and `arrows` move the border.
        `click`, `enter` or `esc` to exit.
                """
            ),
        )

    def _add_scoreboard(self) -> None:

        self.parser.add_argument_group(
            "Scoreboard",
            self.dedent(
                """
    `Single-click` user to highlight and follow.
    `Double-click` user to kick as cheater.
    `Triple-click` user to kick as racist.
    `F7` to change sort column.
                """
            ),
        )

    def _add_logfiles(self) -> None:
        self.parser.add_argument_group(
            "Log Files",
            self.dedent(
                """
    `%(prog)s` writes plaintext messages to `fileno(2)`, and
    colorized messages to `fileno(3)`, when open to a regular file.

        $ %(prog)s 2>x 3>y
                """
            ),
        )

    def main(self) -> None:
        """Command line interface entry point (method)."""

        tf2mon.config = self.config
        tf2mon.options = self.options

        if self.options.con_logfile == self.config["con_logfile"]:
            # Not given on command line; prefix with effective parent.
            # (configured default is a basename relative to `tf2_install_dir`)
            self.options.con_logfile = Path(
                self.config["tf2_install_dir"],
                self.config["con_logfile"],
            )

        if self.options.single_step and not self.options.follow:
            self.parser.error("--no-follow not allowed with --single-step")

        if self.options.list_con_logfile:
            print(self.options.con_logfile)
            self.parser.exit()

        if self.options.trunc_con_logfile:
            Conlog(self.options).trunc()
            logger.info(f"con_logfile {str(self.options.con_logfile)!r} truncated; Exiting.")
            self.parser.exit()

        if self.options.clean_con_logfile:
            Conlog(self.options).clean()
            logger.info(f"con_logfile {str(self.options.con_logfile)!r} cleaned; Exiting.")
            self.parser.exit()

        tf2mon.steam_web_api = SteamWebAPI(str(self.config.get("webapi_key")))

        if self.options.print_steamids:
            Database(self.options.database)
            for steamid in self.options.print_steamids:
                print(tf2mon.steam_web_api.fetch_steamid(steamid))
            self.parser.exit()

        self.monitor.run()


def main(args: list[str] | None = None) -> None:
    """Command line interface entry point (function)."""
    Tf2monCLI(args).main()



"""Table of `SteamPlayer`s."""

from __future__ import annotations

import time
from dataclasses import dataclass

from tf2mon.database import Database, DatabaseTable
from tf2mon.steamid import BOT_STEAMID, SteamID


@dataclass
class SteamPlayer(DatabaseTable):
    """Represent ['response']['players'] element from `ISteamUser.GetPlayerSummaries`."""

    # pylint: disable=too-many-instance-attributes

    __tablename__ = "steamplayers"

    # database columns
    steamid: int  # primary key
    personaname: str = ""
    profileurl: str = ""
    personastate: int = 0
    realname: str = ""
    timecreated: int = 0
    loccountrycode: str = ""
    locstatecode: str = ""
    loccityid: str = ""
    mtime: int = 0

    def __post_init__(self) -> None:

        steamid = SteamID(self.steamid)
        if self.profileurl and self.profileurl == steamid.community_url + "/":
            # for asthetics only; to avoid clutter
            self.profileurl = ""  # indicate long noisy determinable value

        now = int(time.time())

        self.age = 0  # number of days since account was created.
        if self.timecreated:
            self.age = (now - self.timecreated) // 86400
        self.mtime = self.mtime or now

    @classmethod
    def create_table(cls) -> None:
        """Execute create table statement."""

        db = Database()
        assert db

        db.execute(
            f"create table if not exists {cls.__tablename__}"
            """(
                steamid integer primary key,
                personaname text,
                profileurl text,
                personastate integer,
                realname text,
                timecreated integer,
                loccountrycode text,
                locstatecode text,
                loccityid text,
                mtime integer
            )""",
        )
        db.connection.commit()

    @classmethod
    def fetch_steamid(cls, steamid: int) -> SteamPlayer | None:
        """Return `SteamPlayer` for given steamid, else None if not found."""

        db = Database()
        assert db

        db.execute(f"select * from {cls.__tablename__} where steamid=?", (steamid,))
        if row := db.fetchone():
            return cls(*tuple(row))
        return None

    @property
    def is_gamebot(self) -> bool:
        """Return True if this is a legitimate game BOT; not a hacker."""

        return self.steamid == BOT_STEAMID.id  # type: ignore[no-any-return]



from typing import Match

from loguru import logger

import tf2mon
from tf2mon.gameevent import GameEvent
from tf2mon.steamid import parse_steamid
from tf2mon.user import Team


class GameLobbyEvent(GameEvent):

    # tf_lobby_debug
    # "Member[22] [U:1:99999999]  team = TF_GC_TEAM_INVADERS  type = MATCH_PLAYER"

    pattern = r"\s*(?:Member|Pending)\[\d+\] (?P<steamid>\S+)\s+team = (?P<teamname>\w+)"

    def handler(self, match: Match[str]) -> None:

        # this will not be called for games on local server with bots
        # or community servers; only on valve matchmaking servers.

        s_steamid, teamname = match.groups()

        if not (steamid := parse_steamid(s_steamid)):
            return  # invalid

        if teamname == "TF_GC_TEAM_INVADERS":
            team = Team.BLU
        elif teamname == "TF_GC_TEAM_DEFENDERS":
            team = Team.RED
        else:
            logger.critical(f"bad teamname {teamname!r} steamid {steamid}")
            return

        if old_team := tf2mon.users.teams_by_steamid.get(steamid):
            # if we've seen this steamid before...
            if old_team != team:
                # ...and
                logger.warning(f"{steamid.id} change team `{old_team}` to `{team}`")
        else:
            logger.log("ADDLOBBY", f"{team} {steamid.id}")

        #
        tf2mon.users.teams_by_steamid[steamid] = team



"""Monitor Admin console."""

# mypy: ignore-errors
# This file is no longer used.

from loguru import logger

import tf2mon
from tf2mon.player import Player
from tf2mon.regex import Regex


class Admin:
    """Monitor Admin console."""

    regex_list = [
        # stop single-stepping
        Regex("^(c|cont|continue)$", lambda m: tf2mon.SingleStepControl.stop_single_stepping()),
        # stop single-stepping until eof, then single-step again
        Regex("^(r|run|g|go)$", lambda m: tf2mon.SingleStepControl.set_single_step_lineno(0)),
        # dump internals
        Regex("^dump$", lambda m: tf2mon.dump()),
        # pause when logfile reaches `lineno`.
        Regex(
            R"^(b|break|breakpoint)[= ](?P<lineno>\d+)$",
            lambda m: tf2mon.SingleStepControl.set_single_step_lineno(int(m.group("lineno"))),
        ),
        # pause when logfile reaches next line that matches `pattern`.
        Regex(
            "^/(?P<pattern>.*)$",
            lambda m: tf2mon.SingleStepControl.set_single_step_pattern(m.group("pattern")),
        ),
        # kick cheater
        Regex(
            R"^kick[= ](?P<userid>\d+)$",
            lambda m: tf2mon.users.kick_userid(int(m.group("userid")), Player.CHEATER),
        ),
        # kick racist
        Regex(
            R"^kkk[= ](?P<userid>\d+)$",
            lambda m: tf2mon.users.kick_userid(int(m.group("userid")), Player.RACIST),
        ),
        # mark suspect
        Regex(
            R"^suspect[= ](?P<userid>\d+)$",
            lambda m: tf2mon.users.kick_userid(int(m.group("userid")), Player.SUSPECT),
        ),
        # drop to python debugger
        Regex("^PDB$", lambda m: tf2mon.debugger()),
        # deprecated, legacy, support old conlogs.
        Regex(
            "^(TOGGLE-SCOREBOARD|TOGGLE-QUEUES|CHATS-POP|CHATS-POPLEFT|CHATS-CLEAR)$",
            lambda m: logger.debug("ignore"),
        ),
        # The following aren't commands, but other items that may be in
        # the conlog. They don't belong in `gameplay` because they're
        # not generated by tf2.
        # qvalve
        Regex("^(QVALVE) (.*)", lambda m: logger.log(m.group(1), m.group(2))),
    ]



"""Team Fortress 2 Gameplay."""

from .capture import GameCaptureEvent
from .chat import GameChatEvent
from .connected import GameConnectedEvent
from .kill import GameKillEvent
from .lobby import GameLobbyEvent
from .misc import (
    GameHostnameEvent,
    GameLobbyFailedEvent,
    GamePingEvent,
    GameServerEvent,
    GameTeamsSwitchedEvent,
    GameUserSwitchedEvent,
)
from .perk import GamePerkChangeEvent, GamePerkOff1Event, GamePerkOff2Event, GamePerkOnEvent
from .status import GameStatusEvent
from .suicide import GameSuicideEvent

events = [
    GameCaptureEvent(),
    GameServerEvent(),
    GamePingEvent(),
    GameChatEvent(),
    GameKillEvent(),
    GameConnectedEvent(),
    GameSuicideEvent(),
    GameStatusEvent(),
    GameLobbyEvent(),
    GamePerkOnEvent(),
    GamePerkOff1Event(),
    GamePerkOff2Event(),
    GamePerkChangeEvent(),
    GameLobbyFailedEvent(),
    GameTeamsSwitchedEvent(),
    GameUserSwitchedEvent(),
    GameHostnameEvent(),
]



from typing import Match

from loguru import logger

import tf2mon
from tf2mon.gameevent import GameEvent
from tf2mon.steamid import BOT_STEAMID, parse_steamid
from tf2mon.user import UserKey


class GameStatusEvent(GameEvent):

    # status
    # "# userid name                uniqueid            connected ping loss state"
    # "#     29 "Bob"               [U:1:99999999]      01:24       67    0 active"
    # "#    158 "Jones"             [U:1:9999999999]     2:21:27    78    0 active
    # "#      3 "Nobody"            BOT                                     active

    pattern = r'#\s*(?P<s_userid>\d+) "(?P<username>.+)"\s+(?P<steamid>\S+)(?:\s+(?P<elapsed>[\d:]+)\s+(?P<ping>\d+))'

    def handler(self, match: Match[str]) -> None:

        # pylint: disable=too-many-branches
        # pylint: disable=too-many-locals

        s_userid, username, s_steamid, s_elapsed, ping = match.groups()

        tf2mon.ui.notify_operator = False

        if not (steamid := parse_steamid(s_steamid)):
            return  # invalid

        userid = int(s_userid)
        user = None

        if steamid != BOT_STEAMID and (user := tf2mon.users.users_by_steamid.get(steamid)):
            if user.username and user.username != username:
                logger.warning(f"{steamid.id} change username `{user.username}` to `{username}`")
                user.username = username
                if user.player:
                    user.player.track_appearance(username)

            if user.userid and user.userid != userid:
                logger.warning(f"{steamid.id} change userid `{user.userid}` to `{userid}`")
                user.userid = userid

        if not user:
            user = tf2mon.users[UserKey(username)]

        user.dirty = True

        if not user.userid:
            user.userid = userid

        if not user.steamid:
            user.steamid = steamid
            tf2mon.users.users_by_steamid[steamid] = user

        #
        mdy = s_elapsed.split(":")
        if len(mdy) == 2:
            _h, _m, _s = 0, int(mdy[0]), int(mdy[1])
            user.elapsed = (_m * 60) + _s
        elif len(mdy) == 3:
            _h, _m, _s = int(mdy[0]), int(mdy[1]), int(mdy[2])
            user.elapsed = (_h * 3600) + (_m * 60) + _s
        else:
            _h, _m, _s = 0, 0, 0
            user.elapsed = 0

        # hh:mm:ss
        #     0:00
        _ss = f"{_s:02}"
        if not _h:
            _mm = f"{_m:5}"
            user.s_elapsed = _mm + ":" + _ss
        else:
            _hh = f"{_h:2}"
            _mm = f"{_m:02}"
            user.s_elapsed = _hh + ":" + _mm + ":" + _ss

        #
        user.ping = int(ping)
        logger.log("STATUS", user)

        #
        if not user.team and (team := tf2mon.users.teams_by_steamid.get(steamid)):
            user.team = team

        #
        if not user.steamplayer:
            user.vet()



"""A user of the game."""

from __future__ import annotations

import re
import time
from dataclasses import dataclass, field
from enum import Enum
from typing import Iterator, NewType

from loguru import logger

import tf2mon
from tf2mon.chat import Chat
from tf2mon.player import Player
from tf2mon.racist import clean_username
from tf2mon.role import Role
from tf2mon.steamplayer import SteamPlayer

UserKey = NewType("UserKey", str)
WeaponState = NewType("WeaponState", str)


class Team(Enum):
    """Valid teams."""

    RED = 2
    BLU = 3


@dataclass
class UserStats:
    """Snapshot of stats at some point in time."""

    # pylint: disable=too-many-instance-attributes
    user: User | None
    last_killer: User | None
    last_victim: User | None
    role: Role | None
    weapon_state: WeaponState | None
    nkills: int
    ndeaths: int
    kdratio: float
    # timestamp: float


@dataclass
class Kill:
    """A Kill."""

    killer: UserStats  # snapshot of killer at time of kill
    victim: UserStats  # snapshot of victim at time of death
    timestamp: float = field(default=0, init=False)

    def __post_init__(self) -> None:
        """Initialize data and other attributes."""

        if not self.timestamp:
            self.timestamp = time.time()


class User:
    """A user of the game."""

    # pylint: disable=too-many-instance-attributes

    _re_cheater_chats = re.compile(
        "|".join(
            [
                "UNCLETOPIA.COM BEST MODERATED SERVERS! NO BOTS NO CHEATERS!",
                "m4gic m4gic",
            ]
        )
    )

    _max_status_checks = 2

    def __init__(self, username: str) -> None:
        """Create `User`."""

        self.username = username.replace(";", ".")
        self._clean_username = clean_username(self.username)

        # too strict?
        if "\u0e31" in self.username:
            # self.username = self.username.replace("\u0e31", "?")
            logger.warning(f"'\u0e31' in {self.username!r}")

        #
        self.username_upper = username.upper()
        self.userid = 0  # from status command
        self.steamid = None  # from status and tf_lobby_debug commands
        self._team: Team | None = None  # @team.setter
        self.elapsed: int = 0
        self.s_elapsed: str = ""
        self.ping = 0
        self.last_scoreboard_line = ""
        self.dirty = True

        self.n_status_checks = 0
        self.nsnipes = 0
        self.role = Role.unknown
        self.weapon_state: WeaponState | None = None
        self.ncaptures = 0
        self.ndefenses = 0
        self.chats: list[Chat] = []
        self.display_level = ""
        self.selected = False
        self.perk = ""

        #
        self.opponents: dict[UserKey, User] = {}
        self.victims: dict[UserKey, User] = {}
        self.killers: dict[UserKey, User] = {}

        self.last_killer: User | None = None
        self.last_victim: User | None = None

        self.kills: list[Kill] = []
        self.deaths: list[Kill] = []
        self.nkills = 0
        self.ndeaths = 0
        self.kdratio: float = 0

        # "by" as in "lookup by", "for each", "per".
        self.nkills_by_opponent: dict[UserKey, int] = {}
        self.ndeaths_by_opponent: dict[UserKey, int] = {}
        self.kdratio_by_opponent: dict[UserKey, float] = {}
        self.nkills_by_opponent_by_weapon: dict[UserKey, dict[WeaponState, int]] = {}

        # list of non-kill actions performed, like capture/defend.
        self.actions: list[str] = []

        #
        self.steamplayer: SteamPlayer | None = None
        self.age = 0
        self.player: Player | None = None

        # Database `Player`s are keyed by `steamid`. If `self.kick(attr)`
        # is called before steamid is available, spool the work until it is.

        self.pending_attrs: list[str] = []

        # If this looks like a cheater we're tracking by name, mark him to
        # be kicked when his steamid becomes available. Doing this now to
        # notify the operator asap to `TF2MON-PUSH` steamids to us.
        # Careful, this might be a legitimate name-change, not a cheating name-stealer.

        self.cloner: User | None = None  # when this user is being cloned
        self.clonee: User | None = None  # when this user is the name-stealing clone

        self.cheater_chat_seen = False

    def is_cheater_chat(self, chat: Chat) -> bool:
        """Return True first time chat appears to be from a cheater."""

        if self.cheater_chat_seen:
            return False
        self.cheater_chat_seen = bool(self._re_cheater_chats.search(chat.msg))
        return self.cheater_chat_seen

    @property
    def key(self) -> UserKey:
        """Return readable hashable key."""

        return UserKey(f"{self.userid}-{self._clean_username[:15]}")

    @property
    def is_active(self) -> bool:
        """Return True if user is active."""

        return self.n_status_checks < self._max_status_checks

    @property
    def points(self) -> int:
        """Return number of points scored."""

        return self.nkills + self.ncaptures + self.ndefenses

    @property
    def opposing_team(self) -> Team:
        """Return opposing team."""

        return Team.RED if self.team == Team.BLU else Team.BLU

    @property
    def moniker(self) -> str:
        """Return name, optionally including his kill/death ratio."""

        if not tf2mon.ShowKDControl.value:
            return self._clean_username

        # pylint: disable=consider-using-f-string

        if self.ndeaths < 2:
            return "{!r} ({}/{})".format(self._clean_username, self.nkills, self.ndeaths)

        return "{!r} ({}/{}={:.1f})".format(
            self._clean_username, self.nkills, self.ndeaths, self.nkills / self.ndeaths
        )

    def duel_as_str(self, opponent: User, formatted: bool = False) -> str:
        """Return string showing win/loss record against `opponent`."""

        nkills = self.nkills_by_opponent.get(opponent.key, 0)
        ndeaths = self.ndeaths_by_opponent.get(opponent.key, 0)
        return f"{nkills:2} and {ndeaths:2}" if formatted else f"{nkills} and {ndeaths}"

    def __repr__(self) -> str:

        team = f"{self.team.name}:" if self.team else ""
        return f"{team}{self.userid}={self.username!r}"

    def snap_stats(self) -> UserStats:
        """Take and return a snapshot of current stats."""

        return UserStats(
            self,
            self.last_killer,
            self.last_victim,
            self.role,
            self.weapon_state,
            self.nkills,
            self.ndeaths,
            self.kdratio,
        )

    def format_user_stats(self) -> Iterator[str]:
        """Return most recent kill and most recent death in reverse chronological order."""

        kills: list[Kill] = []

        if len(self.kills) > 0:
            kills.append(self.kills[-1])
        if len(self.deaths) > 0:
            kills.append(self.deaths[-1])

        for kill in sorted(kills, key=lambda x: -x.timestamp):
            if self == kill.killer.user:
                opponent = kill.victim.user
                label = "[last-Victim]"
            else:
                opponent = kill.killer.user
                label = "[last-Killer]"

            assert opponent

            yield " ".join(
                [
                    time.strftime("%T", time.localtime(kill.timestamp)),
                    label,
                    self.duel_as_str(opponent, formatted=True),
                    "vs",
                    f"{opponent.username:20.20}",
                    opponent.weapon_state or "",
                ]
            )

    @property
    def team(self) -> Team | None:
        """Return user's `Team`."""
        return self._team

    @team.setter
    def team(self, team: Team | str) -> None:
        """Assign this user to `team`."""

        if isinstance(team, str):
            if team == Team.RED.name:
                team = Team.RED
            elif team == Team.BLU.name:
                team = Team.BLU
            else:
                logger.critical(f"bad team {team!r}")
                return

        if not self.team:
            logger.info(f"{self} joins {team}")

        elif self.team != team:
            logger.debug(f"{self} change from {self.team} to {team}")

        self._team = team
        self.dirty = True

        # assign any unassigned opponents

        for opponent in self.opponents.values():
            if not opponent.team:
                opponent.team = self.opposing_team
            if not self.team:
                self.team = opponent.opposing_team

    def vet(self) -> None:
        """Vet this player, whose `steamid` has just been obtained."""

        assert self.steamid
        self.dirty = True

        self.steamplayer = tf2mon.steam_web_api.fetch_steamid(self.steamid.id)
        if self.steamplayer.is_gamebot:
            self.steamplayer.personaname = self.username
            self.pending_attrs = []
            return

        logger.log("SteamPlayer", f"{self} SteamPlayer={self.steamplayer}")
        self.age = self.steamplayer.age

        # known hacker?
        self.player = Player.fetch_steamid(self.steamid.id)
        if self.player:
            # logger.log("Player", self.player.astuple())
            self.player.setattrs(self.pending_attrs)
            self.player.track_appearance(self.username)
            tf2mon.ui.show_player_intel(self.player)
            # bobo1
            self.display_level = self.player.display_level
            # logger.log(self.display_level, f"{self._clean_username!r} is here")
            # bobo2
            self.pending_attrs = None
            if self.player.is_banned:
                self.do_kick()
            return
        logger.trace(f"{self} is not a known hacker")

        # Have we tried to kick them, but had to spool the work because
        # `steamid` wasn't available yet?
        if self.pending_attrs:
            self.player = Player.new_player(self.steamid.id, self.pending_attrs, self.username)
            # bobo1
            self.display_level = self.player.display_level
            logger.log(self.display_level, f"{self} created {self.player}")
            # bobo2
            self.pending_attrs = None
            if self.player.is_banned:
                self.do_kick()

    def kick(self, attr: str) -> None:
        """Kick this user."""

        if not self.steamid:
            # postpone work until steamid available
            self.pending_attrs.append(attr)
            self.display_level = attr.upper()
            logger.log(self.display_level, f"{self} needs steamid, Press KP_DOWNARROW to PUSH")
            tf2mon.ui.notify_operator = True
            tf2mon.ui.sound_alarm = True
            return

        if self.player:
            if not getattr(self.player, attr):
                self.player.setattrs([attr])
                self.player.upsert()
                logger.log(self.display_level, f"{self} added {attr} to {self.player}")
            else:
                logger.info(f"{self} player {self.player} already {attr}")
        else:
            self.player = Player.new_player(
                self.steamid.id, [attr] + self.pending_attrs, self.username
            )
            self.display_level = self.player.display_level
            logger.log(self.display_level, f"{self} created {self.player}")
            self.pending_attrs = None

        if self.player.is_banned:
            self.do_kick()

    def do_kick(self) -> None:
        """Work."""

        assert self.player

        # msg = f"say {tf2mon.APPNAME} ALERT: "
        msg = "say ALERT: "
        if self.player.racist or self.player._racist:  # noqa
            msg += f"RACIST {self._clean_username!r}"
        elif self.clonee:
            msg += f"NAME-STEALER {self.username!r}"
        else:
            msg += f"CHEATER {self.username!r}"

        msg += " is here"
        cmd = f"CALLVOTE KICK {self.userid}"
        msg += f", {cmd}"

        tf2mon.KicksControl.push(msg)
        tf2mon.KicksControl.push(cmd)



"""Circular list that returns next item with each access."""

from typing import Any


class Cycle:
    """Circular list that returns next item with each access."""

    def __init__(self, name: str, values: Any) -> None:
        """Create new `Cycle`.

        Args:
            name:   name of instance.
            values: list of values to cycle through.
        """

        super().__init__()

        self.name = name
        assert values
        self._values = list(values)
        _len = len(self._values)
        assert _len
        self._max = _len - 1
        self._idx = 0

    def __setitem__(self, _index: int, value: Any) -> None:
        if value not in self._values:
            raise KeyError(f"value {value!r} not in {self.name!r}")
        while self._values[self._idx] != value:
            _ = self.next

    def start(self, value: Any) -> None:
        """Set starting `value`."""
        self.__setitem__(0, value)  # pylint: disable=unnecessary-dunder-call

    def __getitem__(self, index: int) -> Any:
        return self._values[index]

    def __repr__(self) -> str:
        return str(self._values)

    @property
    def prev(self) -> Any:
        """Return previous value."""
        if self._max:
            self._idx = self._max if self._idx == 0 else self._idx - 1
        return self._values[self._idx]

    @property
    def next(self) -> Any:
        """Return next value."""
        if self._max:
            self._idx = 0 if self._idx == self._max else self._idx + 1
        return self._values[self._idx]

    cycle = next

    @property
    def value(self) -> Any:
        """Return current value."""
        return self._values[self._idx]



from typing import Match

from loguru import logger

import tf2mon
from tf2mon.gameevent import GameEvent
from tf2mon.role import Role, get_role_weapon_state
from tf2mon.spammer import Spammer
from tf2mon.user import Kill, UserKey, WeaponState


class GameKillEvent(GameEvent):

    pattern = r"(?P<killer>.*) killed (?P<victim>.*) with (?P<weapon>.*)\.(?P<crit> \(crit\))?$"
    spammer = Spammer()

    def handler(self, match: Match[str]) -> None:

        # pylint: disable=too-many-branches
        # pylint: disable=too-many-statements

        s_killer, s_victim, weapon, s_crit = match.groups()

        killer = tf2mon.users[UserKey(s_killer)]
        victim = tf2mon.users[UserKey(s_victim)]

        killer.last_victim = victim
        victim.last_killer = killer

        # do most calculations now (once);
        # to avoid calculating when rendering scoreboard (often).

        killer.opponents[victim.key] = victim
        victim.opponents[killer.key] = killer

        killer.victims[victim.key] = victim
        victim.killers[killer.key] = killer

        # totals ---------------------------------------------------------------

        killer.nkills += 1
        victim.ndeaths += 1
        killer.dirty = True
        victim.dirty = True

        _k = killer.nkills
        _d = killer.ndeaths
        killer.kdratio = float(_k) if not _d else _k / _d

        _k = victim.nkills
        _d = victim.ndeaths
        victim.kdratio = float(_k) if not _d else _k / _d

        # subtotals by opponent-------------------------------------------------

        if victim.key not in killer.nkills_by_opponent:
            killer.nkills_by_opponent[victim.key] = 0
        killer.nkills_by_opponent[victim.key] += 1

        if killer.key not in victim.ndeaths_by_opponent:
            victim.ndeaths_by_opponent[killer.key] = 0
        victim.ndeaths_by_opponent[killer.key] += 1

        _k = killer.nkills_by_opponent.get(victim.key, 0)
        _d = killer.ndeaths_by_opponent.get(victim.key, 0)
        killer.kdratio_by_opponent[victim.key] = float(_k) if not _d else _k / _d

        _k = victim.nkills_by_opponent.get(killer.key, 0)
        _d = victim.ndeaths_by_opponent.get(killer.key, 0)
        victim.kdratio_by_opponent[killer.key] = float(_k) if not _d else _k / _d

        # subtotal opponents by weapon_state -----------------------------------

        crit = bool(s_crit)
        role, weapon_state = get_role_weapon_state(killer.role, weapon, crit, killer.perk)
        if role:
            killer.role = role
            killer.weapon_state = WeaponState(weapon_state)
            if killer.role == Role.sniper:
                killer.nsnipes += 1
        else:
            role = killer.role
            if weapon not in ("player", "world"):
                logger.error(f"cannot map {weapon} for {killer} {role}")

        kill = Kill(killer.snap_stats(), victim.snap_stats())
        killer.kills.append(kill)
        victim.deaths.append(kill)

        if not role and weapon not in ("player", "world"):
            logger.error(f"cannot map {weapon} for {killer} {role}")

        if victim.key not in killer.nkills_by_opponent_by_weapon:
            # contains a hash of counts by weapon_state
            killer.nkills_by_opponent_by_weapon[victim.key] = {}
        assert killer.weapon_state
        if weapon_state not in killer.nkills_by_opponent_by_weapon[victim.key]:
            killer.nkills_by_opponent_by_weapon[victim.key][killer.weapon_state] = 0
        killer.nkills_by_opponent_by_weapon[victim.key][killer.weapon_state] += 1

        #
        level = "KILL"
        if killer.team:
            level += killer.team.name

        logger.log(
            level,
            "killer {!r} victim {!r} weapon {!r}",
            killer.moniker,
            victim.moniker,
            weapon_state,
        )

        if killer == tf2mon.users.me:
            if tf2mon.TauntFlagControl.value:
                self.spammer.taunt(victim, weapon, s_crit)
        elif victim == tf2mon.users.me and tf2mon.ThroeFlagControl.value:
            self.spammer.throe(killer, weapon, s_crit)

        if not victim.team and killer.team:
            victim.team = killer.opposing_team
        elif not killer.team and victim.team:
            killer.team = victim.opposing_team



"""Collection of `User` objects."""

import re
from typing import Iterator

from fuzzywuzzy import fuzz  # type: ignore
from loguru import logger

import tf2mon
from tf2mon.player import Player
from tf2mon.racist import is_racist_text
from tf2mon.user import Team, User, UserKey


class Users:
    """Collection of `User`s."""

    def __init__(self) -> None:
        """Initialize collection of `User`s."""

        self.users_by_username: dict[UserKey, User] = {}
        self.users_by_steamid: dict[int, User] = {}
        self.teams_by_steamid: dict[int, Team] = {}
        self.me: User
        self.my: User
        self._max_status_checks = 2

    def __getitem__(self, username: UserKey) -> User:
        """Create user `username` if non-existent, and return user `username`."""

        username = UserKey(username.replace(";", "."))

        if not (user := self.users_by_username.get(username)):
            user = User(username)
            self.users_by_username[UserKey(user.username)] = user
            logger.log("ADDUSER", user)

            if self._is_cheater_name(user):
                user.kick(Player.CHEATER)

            if is_racist_text(username):
                user.kick(Player.RACIST)

        # reset inactivity counter
        if not user.is_active:
            logger.debug(f"Active again {user}")
        user.n_status_checks = 0
        return user

    def active_users(self) -> Iterator[User]:
        """Yield active users (unsorted)."""

        yield from [x for x in self.users_by_username.values() if x.is_active]

    def sorted(self) -> Iterator[User]:
        """Yield active users in sort order."""

        yield from sorted(
            self.active_users(),
            key=tf2mon.SortOrderControl.value,
        )

    def kick_userid(self, userid: int, attr: str) -> None:
        """Kick `userid` reason `attr`."""

        users = [x for x in self.users_by_username.values() if x.userid == userid]
        if len(users) == 1:
            users[0].kick(attr)
        else:
            logger.error(f"bad userid {userid!r}")

    def kick_my_last_killer(self, attr: str) -> None:
        """Kick the last user who killed the operator."""

        if self.my.last_killer:
            self.my.last_killer.kick(attr)
        else:
            logger.warning("no last killer")

    def check_status(self) -> None:
        """Delete users that appear to have left the game.

        We are called in response to TF2MON-PUSH; which may have came long
        before `status` finished sending everything; and we don't have a
        way to detect when `status` finishes; so we'll never be completely
        current. That's why `_max_status_checks` should be at least 2 or 3.
        """

        for user in [x for x in self.users_by_username.values() if x != self.me]:
            was_active = user.is_active
            user.n_status_checks += 1
            if was_active and not user.is_active:
                logger.log("INACTIVE", user)

    def switch_teams(self) -> None:
        """Switch teams."""

        for user in self.active_users():
            user.team = user.opposing_team

    re_cheater_names = re.compile(
        "|".join(
            [
                r"^(\(\d+\))?Sydney",
                r"Swonk Bot",
                r"spoooky braaaap",
                r"Bot Removal Service",
                r"^\[g0tb0t\]Church-of-myg0t",
            ]
        )
    )

    def _is_cheater_name(self, user: User) -> bool:

        if self.re_cheater_names.search(user.username):
            return True

        for _user in [x for x in self.active_users() if x.steamplayer and not x.player]:
            ratio = fuzz.ratio(user.username, _user.username)
            if ratio > 80:
                logger.log("FUZZ", f"ratio {ratio} `{user.username}` vs `{_user.username}`")
                # Careful, this might be a legitimate name-change, not a cheating name-stealer.
                _user.cloner = user  # point the original user to the clone
                user.clonee = _user  # point the clone to the original user
                return True

        return False



"""Cycle contents of User Panel."""

from enum import Enum
from typing import Match

import tf2mon
from tf2mon.control import CycleControl
from tf2mon.cycle import Cycle


class UserPanelControl(CycleControl):
    """Cycle contents of User Panel."""

    name = "TOGGLE-USER-PANEL"
    enum = Enum("enum", "AUTO DUELS KICKS SPAMS")
    cycle = Cycle("_t_user_panel", enum)
    items = {x: x for x in list(enum)}

    def handler(self, _match: Match[str] | None) -> None:
        _ = self.cycle.next
        tf2mon.ui.update_display()



"""User Interface."""

import contextlib
import curses
import os
import sys

import libcurses
from libcurses.bw import BorderedWindow
from loguru import logger

import tf2mon
from tf2mon.baselayout import BaseLayout
from tf2mon.chat import Chat
from tf2mon.player import Player
from tf2mon.scoreboard import Scoreboard
from tf2mon.user import Team, User

# from playsound import playsound


class UI:
    """User Interface."""

    # pylint: disable=too-many-instance-attributes

    def __init__(self, win: curses.window):
        """Initialize User Interface."""

        self.notify_operator = False
        self.sound_alarm = False

        # create empty grid
        self.grid = libcurses.Grid(win)

        self.scoreboard: Scoreboard
        self.logsink: libcurses.LogSink
        # self.colormap: dict[str, int]
        # self.layout: BaseLayout | None = None  # set by `build_grid`.
        self.layout: BaseLayout  # set by `build_grid`.

        # `register_builder` 1) calls `build_grid` and 2) configures
        # `KEY_RESIZE` to call it again each time that event occurs.
        self.grid.register_builder(self.build_grid)
        # assert self.layout

        # begin logging to curses window, too.
        self.logsink = libcurses.LogSink(self.layout.logger_win)

        # map of `loguru-level-name` to `curses-color/attr`.
        self.colormap = libcurses.get_colormap()

        #
        self.scoreboard = Scoreboard(
            self.layout.scorewin_blu,
            self.colormap[Team.BLU.name],
            self.layout.scorewin_red,
            self.colormap[Team.RED.name],
        )

        self.popup_win: BorderedWindow | None = None

    def build_grid(self) -> None:
        """Add boxes to grid.

        Called at init, on KEY_RESIZE events, and when layout changes.
        """

        klass = tf2mon.GridLayoutControl.value
        try:
            self.layout = klass(self.grid)
        except AssertionError:
            curses.endwin()
            msg = "Terminal too small; try `Maximize` and `Ctrl+Minus`."
            logger.error(msg)
            if not os.isatty(sys.stderr.fileno()):
                print(f"ERROR: {msg}")
            sys.exit(1)

        if os.isatty(sys.stderr.fileno()):
            os.close(sys.stderr.fileno())

        for attr in ["logger_win", "status_win", "cmdline_win"]:
            if not hasattr(self.layout, attr):
                raise NotImplementedError(f"undefined `{attr}` in {self.layout}.")

        for win in [
            self.layout.chatwin_blu,
            self.layout.chatwin_red,
            self.layout.logger_win,
        ]:
            if win:
                win.scrollok(True)

        for win in [
            self.layout.scorewin_blu,
            self.layout.scorewin_red,
            self.layout.user_win,
            self.layout.kicks_win,
            self.layout.spams_win,
            self.layout.duels_win,
            self.layout.status_win,
        ]:
            if win:
                win.scrollok(False)

        if self.layout.cmdline_win:
            self.layout.cmdline_win.scrollok(True)
            self.layout.cmdline_win.keypad(True)

        if (
            tf2mon.ui is not None
            and self.layout.chatwin_blu is not None
            and self.layout.chatwin_red is not None
        ):
            tf2mon.ChatsControl.refresh()
        self.grid.redraw()

    def getline(self, prompt: str | None = None) -> str | None:
        """Read and return next line from keyboard."""

        self.layout.cmdline_win.erase()
        if prompt:
            self.layout.cmdline_win.addstr(0, 0, prompt)
        return libcurses.getline(self.layout.cmdline_win)

    def update_display(self) -> None:
        """Update display."""

        if self.sound_alarm:
            self.sound_alarm = False
            if (
                tf2mon.conlog and tf2mon.conlog.is_eof
            ):  # don't do this when replaying logfile from start
                ...
                # playsound('/usr/share/sounds/sound-icons/prompt.wav')
                # playsound('/usr/share/sounds/sound-icons/cembalo-10.wav')
                # curses.flash()
                # curses.beep()

        self.refresh_kicks()
        self.refresh_spams()
        self.refresh_duels(tf2mon.users.me)
        self.refresh_user(tf2mon.users.me)
        # chatwin_blu and chatwin_red are rendered from gameplay/_playerchat
        self.scoreboard.refresh()
        self.show_status()

        if self.popup_win:
            self.popup_win.refresh()

    def refresh_kicks(self) -> None:
        """Refresh kicks panel."""

        if self.layout.kicks_win:
            self._show_lines(
                "KICKS",
                list(reversed(tf2mon.KicksControl.msgs)),
                self.layout.kicks_win,
            )

        if self.layout.user_win:
            self.refresh_user(tf2mon.users.me)

    def refresh_spams(self) -> None:
        """Refresh spams panel."""

        if self.layout.spams_win:
            self._show_lines(
                "SPAMS",
                list(reversed(tf2mon.SpamsControl.msgs)),
                self.layout.spams_win,
            )

        if self.layout.user_win:
            self.refresh_user(tf2mon.users.me)

    def refresh_duels(self, user: User) -> None:
        """Refresh duels panel."""

        if self.layout.duels_win:
            self._show_lines("user", self._format_duels(user), self.layout.duels_win)

        if self.layout.user_win:
            self.refresh_user(tf2mon.users.me)

    def refresh_user(self, user: User) -> None:
        """Refresh user panel."""

        panel = tf2mon.UserPanelControl
        kicks = tf2mon.KicksControl
        spams = tf2mon.SpamsControl

        if self.layout.user_win:
            if panel.value == panel.enum.KICKS or (
                panel.value == panel.enum.AUTO and kicks.msgs
            ):
                self._show_lines(
                    "KICKS",
                    list(reversed(kicks.msgs)) if kicks.msgs else ["No Kicks"],
                    self.layout.user_win,
                )
            #
            elif panel.value == panel.enum.SPAMS or (
                panel.value == panel.enum.AUTO and spams.msgs
            ):
                self._show_lines(
                    "SPAMS",
                    list(reversed(spams.msgs)) if spams.msgs else ["No Spams"],
                    self.layout.user_win,
                )
            #
            else:
                self._show_lines("user", self._format_duels(user), self.layout.user_win)

            self.layout.user_win.noutrefresh()

    def user_color(self, user: User, color: int) -> int:
        """Return `color` to display `user` in scoreboard."""

        if user.display_level:
            color = self.colormap[user.display_level]

        if user == tf2mon.users.my.last_killer:
            color |= curses.A_BOLD | curses.A_ITALIC

        if user == tf2mon.users.my.last_victim:
            color |= curses.A_BOLD

        if user.selected:
            color |= curses.A_REVERSE

        if not user.team:
            color |= curses.A_UNDERLINE

        if user.cloner:
            color |= curses.A_ITALIC

        return color

    def show_chat(self, chat: Chat) -> None:
        """Display (append) `chat` in appropriate team window."""

        win = None
        if chat.user.team == Team.RED:
            win = self.layout.chatwin_red
        if not win:
            win = self.layout.chatwin_blu  # unassigned, or RED in shared window.
        if not win:
            return  # not showing chats

        user = chat.user
        color = self.colormap[user.team.name if user.team else "user"]
        if chat.teamflag:
            color |= curses.A_UNDERLINE
        user_color = self.user_color(user, color)

        leader = f"{chat.s_timestamp}: {chat.user.username:20.20}: "
        if sum(win.getyx()):
            win.addch("\n")
        win.addstr(leader + chat.msg, user_color)

        if tf2mon.ShowKillsControl.value and chat.msg != "/rtd":
            # self._show_last_duels(win, leader, user, user_color)
            indent = " " * 15
            for line in chat.user.format_user_stats():
                win.addstr(f"\n{leader}{indent}{line}")

        win.noutrefresh()

    def show_journal(self, level: str, line: str) -> None:
        """Display `line` in some pseudo "journal" window."""

        assert self.layout
        if (win := self.layout.chatwin_blu) is None:
            return

        if sum(win.getyx()):
            win.addch("\n")
        win.addstr(line, self.colormap[level])
        win.noutrefresh()

    def popup(self, level: str, text: str) -> None:
        """Display `text` in a popup window."""

        self.popup_win = BorderedWindow(
            self.grid.nlines - 10,
            80,
            self.grid.begin_y + 5,
            self.grid.begin_x + 20,
        )
        # self.popup_win.box()
        # self.popup_win.scrollok(True)
        self.popup_win.w.addstr(text, self.colormap[level])
        self.popup_win.refresh()

    def show_player_intel(self, player: Player) -> None:
        """Display what we know about `player`."""

        level = player.display_level
        leader = f"{level}: {player.steamid}"

        self.show_journal(
            level,
            f"{leader}: name: `{player.last_name}`",
        )

        for alias in [x for x in player.aliases if x != player.last_name]:
            self.show_journal(
                level,
                f"{leader}: alias: `{alias}`",
            )

        self.show_journal(
            level,
            # pylint: disable=protected-access
            f"{leader}: prev={player.s_prev_time} {player._s_prev_time}",
        )

        self.show_journal(
            level,
            f"{leader}: attrs={[x for x in player.getattrs() if x]}",
        )

    def _format_duels(self, user: User) -> list[str]:

        lines: list[str] = []
        indent = " " * 12  # 12=len("99 and 99 vs")

        lines.append("Duels:")
        for opponent in [x for x in user.opponents.values() if x.is_active]:
            lines.append(f"{user.duel_as_str(opponent, True)} vs {opponent.moniker}")

            if opponent.key in user.nkills_by_opponent_by_weapon:
                for weapon, count in user.nkills_by_opponent_by_weapon[opponent.key].items():
                    lines.append(f"{indent} K {count:2} {weapon}")

            if user.key in opponent.nkills_by_opponent_by_weapon:
                for weapon, count in opponent.nkills_by_opponent_by_weapon[user.key].items():
                    lines.append(f"{indent} D {count:2} {weapon}")

        return lines

    def show_status(self) -> None:
        """Update status line."""

        line = tf2mon.controller.get_status_line() + f" UID={tf2mon.users.my.userid}"

        try:
            self.layout.status_win.addstr(
                0, 0, line, self.colormap["NOTIFY" if self.notify_operator else "STATUS"]
            )
            self.layout.status_win.clrtoeol()
        except curses.error:
            pass
        self.layout.status_win.noutrefresh()

    def _show_lines(self, level: str, lines: list[str], win: curses.window) -> None:

        win.erase()

        with contextlib.suppress(curses.error):
            for line in lines:
                win.addstr(line + "\n", self.colormap[level])
        win.noutrefresh()



"""Docstring."""

from loguru import logger
from steam.steamid import SteamID  # type: ignore[import-untyped]

__all__ = ["SteamID"]

BOT_STEAMID = SteamID(1)


def parse_steamid(s_steamid: str) -> SteamID:
    """Parse and return `SteamID` from STATUS and LOBBY lines.

    Return `None` if invalid, or `BOT_STEAMID` if gamebot.
    """

    if s_steamid == "BOT":
        return BOT_STEAMID

    steamid = SteamID(s_steamid)
    if not steamid.is_valid():
        logger.error(f"invalid steamid {s_steamid}")
        return None

    return steamid



"""Player `Chat` message."""

from __future__ import annotations

import time
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from tf2mon.user import User, UserStats


@dataclass
class Chat:
    """Player `Chat` message."""

    user: User
    teamflag: bool
    msg: str
    timestamp: float | None = field(default=None)
    #
    s_timestamp: str | None = field(default=None, init=False)
    stats: UserStats | None = field(default=None, init=False)

    def __post_init__(self) -> None:
        """Initialize data and other attributes."""

        if not self.timestamp:
            self.timestamp = time.time()

        self.s_timestamp = time.strftime("%T", time.localtime(self.timestamp))
        # _dt = datetime.datetime.fromtimestamp(self.timestamp)
        # self.s_timestamp = _dt.strftime("%T.%f")  # [:-4]

        self.stats = self.user.snap_stats()



"""Miscelleneous controls."""

from typing import Match

import tf2mon
from tf2mon.control import BoolControl, Control
from tf2mon.cycle import Cycle


class DebugFlagControl(BoolControl):
    """Enable/disable debug (use `ECHO` or `SAY`)."""

    name = "TOGGLE-DEBUG"
    cycle = Cycle("debug", [False, True])


class ShowDebugControl(Control):
    """Show debugging."""

    name = "SHOW-DEBUG"

    def handler(self, _match: Match[str] | None) -> None:
        tf2mon.ui.show_journal("help", " Grid ".center(80, "-"))
        tf2mon.ui.show_journal("help", str(tf2mon.ui.grid))
        for box in tf2mon.ui.grid.boxes:
            tf2mon.ui.show_journal("help", tf2mon.ui.grid.winyx(box))


class TauntFlagControl(BoolControl):
    """Enable/disable `Taunt` messages."""

    name = "TOGGLE-TAUNT"
    cycle = Cycle("taunt", [False, True])


class ThroeFlagControl(BoolControl):
    """Enable/disable `Throe` messages."""

    name = "TOGGLE-THROE"
    cycle = Cycle("throe", [True, False])


class ShowKDControl(BoolControl):
    """Include `Kill/Death ratio` in `User.moniker`."""

    name = "TOGGLE-KD"
    cycle = Cycle("kd", [False, True])


class ShowKillsControl(BoolControl):
    """Display kills in journal window."""

    name = "TOGGLE-SHOW-KILLS"
    cycle = Cycle("kills", [False, True])

    def handler(self, _match: Match[str] | None) -> None:
        super().handler(_match)
        tf2mon.ChatsControl.refresh()


class ShowPerksControl(Control):
    """Display perks in journal window."""

    name = "SHOW-PERKS"

    def handler(self, _match: Match[str] | None) -> None:
        tf2mon.ui.show_journal("help", " Perks ".center(80, "-"))
        for user in [x for x in tf2mon.users.active_users() if x.perk]:
            tf2mon.ui.show_journal("help", f"{user!r:25} {user.perk}")


class JoinOtherTeamControl(Control):
    """Join Other Team."""

    name = "SWITCH-MY-TEAM"

    def handler(self, _match: Match[str] | None) -> None:
        if self.toggling_enabled():
            tf2mon.users.my.team = tf2mon.users.my.opposing_team
            tf2mon.ui.update_display()

    def status(self) -> str:
        assert tf2mon.users.my.team
        return tf2mon.users.my.team.name  # if tf2mon.users.my.team else "blu"


class ClearQueuesControl(Control):
    """Clear kicks and spams queues."""

    name = "CLEAR-QUEUES"
    action = "tf2mon_clear_queues"

    def handler(self, _match: Match[str] | None) -> None:
        """Handle event."""

        tf2mon.KicksControl.clear()
        tf2mon.ui.refresh_kicks()
        tf2mon.SpamsControl.clear()
        tf2mon.ui.refresh_spams()


class PushControl(Control):
    """Push `steamids` from game to monitor."""

    name = "PUSH"
    action = "tf2mon_push"



"""Team Fortress 2 Console Monitor."""

import curses
from argparse import Namespace
from pprint import pformat
from typing import Any

from loguru import logger

from tf2mon.conlog import Conlog
from tf2mon.controller import Controller
from tf2mon.steamweb import SteamWebAPI
from tf2mon.ui import UI
from tf2mon.user import Team, UserKey
from tf2mon.users import Users

config: dict[str, Any] = {}
conlog: Conlog | None = None
options: Namespace
steam_web_api: SteamWebAPI
ui: UI
users: Users

from tf2mon.controls.chats import ChatsControl as _ChatsControl  # noqa
from tf2mon.controls.chats import ClearChatsControl as _ClearChatsControl  # noqa
from tf2mon.controls.chats import RefreshChatsControl as _RefreshChatsControl  # noqa
from tf2mon.controls.gridlayout import GridLayoutControl as _GridLayoutControl  # noqa
from tf2mon.controls.help import HelpControl as _HelpControl  # noqa
from tf2mon.controls.help import MotdControl as _MotdControl  # noqa
from tf2mon.controls.kicklast import KickLastCheaterControl as _KickLastCheaterControl  # noqa
from tf2mon.controls.kicklast import KickLastRacistControl as _KickLastRacistControl  # noqa
from tf2mon.controls.kicklast import KickLastSuspectControl as _KickLastSuspectControl  # noqa
from tf2mon.controls.kicks import KicksClearControl as _KicksClearControl  # noqa
from tf2mon.controls.kicks import KicksControl as _KicksControl  # noqa
from tf2mon.controls.kicks import KicksPopControl as _KicksPopControl  # noqa
from tf2mon.controls.kicks import KicksPopleftControl as _KicksPopleftControl  # noqa
from tf2mon.controls.logger import LogLevelControl as _LogLevelControl  # noqa
from tf2mon.controls.logger import LogLocationControl as _LogLocationControl  # noqa
from tf2mon.controls.logger import ResetPaddingControl as _ResetPaddingControl  # noqa
from tf2mon.controls.misc import ClearQueuesControl as _ClearQueuesControl  # noqa
from tf2mon.controls.misc import DebugFlagControl as _DebugFlagControl  # noqa
from tf2mon.controls.misc import JoinOtherTeamControl as _JoinOtherTeamControl  # noqa
from tf2mon.controls.misc import PushControl as _PushControl  # noqa
from tf2mon.controls.misc import ShowDebugControl as _ShowDebugControl  # noqa
from tf2mon.controls.misc import ShowKDControl as _ShowKDControl  # noqa
from tf2mon.controls.misc import ShowKillsControl as _ShowKillsControl  # noqa
from tf2mon.controls.misc import ShowPerksControl as _ShowPerksControl  # noqa
from tf2mon.controls.misc import TauntFlagControl as _TauntFlagControl  # noqa
from tf2mon.controls.misc import ThroeFlagControl as _ThroeFlagControl  # noqa
from tf2mon.controls.msgqueues import DisplayFileControl as _DisplayFileControl  # noqa
from tf2mon.controls.msgqueues import MsgQueuesControl as _MsgQueuesControl  # noqa
from tf2mon.controls.singlestep import SingleStepControl as _SingleStepControl  # noqa
from tf2mon.controls.singlestep import SingleStepStartControl as _SingleStepStartControl  # noqa
from tf2mon.controls.sortorder import SortOrderControl as _SortOrderControl  # noqa
from tf2mon.controls.spams import SpamsClearControl as _SpamsClearControl  # noqa
from tf2mon.controls.spams import SpamsControl as _SpamsControl  # noqa
from tf2mon.controls.spams import SpamsPopControl as _SpamsPopControl  # noqa
from tf2mon.controls.spams import SpamsPopleftControl as _SpamsPopleftControl  # noqa
from tf2mon.controls.userpanel import UserPanelControl as _UserPanelControl  # noqa

controller = Controller(
    [
        ChatsControl := _ChatsControl(),
        ClearChatsControl := _ClearChatsControl(),
        ClearQueuesControl := _ClearQueuesControl(),
        DebugFlagControl := _DebugFlagControl(),
        DisplayFileControl := _DisplayFileControl(),
        GridLayoutControl := _GridLayoutControl(),
        HelpControl := _HelpControl(),
        JoinOtherTeamControl := _JoinOtherTeamControl(),
        KickLastCheaterControl := _KickLastCheaterControl(),
        KickLastRacistControl := _KickLastRacistControl(),
        KickLastSuspectControl := _KickLastSuspectControl(),
        KicksClearControl := _KicksClearControl(),
        KicksControl := _KicksControl(),
        KicksPopControl := _KicksPopControl(),
        KicksPopleftControl := _KicksPopleftControl(),
        LogLevelControl := _LogLevelControl(),
        LogLocationControl := _LogLocationControl(),
        MotdControl := _MotdControl(),
        MsgQueuesControl := _MsgQueuesControl(),
        PushControl := _PushControl(),
        RefreshChatsControl := _RefreshChatsControl(),
        ResetPaddingControl := _ResetPaddingControl(),
        ShowDebugControl := _ShowDebugControl(),
        ShowKDControl := _ShowKDControl(),
        ShowKillsControl := _ShowKillsControl(),
        ShowPerksControl := _ShowPerksControl(),
        SingleStepControl := _SingleStepControl(),
        SingleStepStartControl := _SingleStepStartControl(),
        SortOrderControl := _SortOrderControl(),
        SpamsClearControl := _SpamsClearControl(),
        SpamsControl := _SpamsControl(),
        SpamsPopControl := _SpamsPopControl(),
        SpamsPopleftControl := _SpamsPopleftControl(),
        TauntFlagControl := _TauntFlagControl(),
        ThroeFlagControl := _ThroeFlagControl(),
        UserPanelControl := _UserPanelControl(),
    ]
)

# ordered for rendering `--help`
controller.bind(HelpControl, "F1")
controller.bind(MotdControl, "Ctrl+F1")
controller.bind(DisplayFileControl, "Shift+F1")
controller.bind(DebugFlagControl, "F2")
controller.bind(TauntFlagControl, "F3")
controller.bind(ThroeFlagControl, "Shift+F3")
controller.bind(ShowKDControl, "F4")
controller.bind(ShowKillsControl, "Shift+F4")
controller.bind(UserPanelControl, "F5")
controller.bind(ShowPerksControl, "Shift+F5")
controller.bind(JoinOtherTeamControl, "F6")
controller.bind(SortOrderControl, "F7")
controller.bind(LogLocationControl, "F8")
controller.bind(ResetPaddingControl, "Ctrl+F8")
controller.bind(LogLevelControl, "Shift+F8")
controller.bind(GridLayoutControl, "F9")
controller.bind(RefreshChatsControl, "Ctrl+F9")
controller.bind(ClearChatsControl, "Shift+F9")
controller.bind(ShowDebugControl, "KP_INS")
controller.bind(SingleStepStartControl, "KP_DEL")
controller.bind(KickLastCheaterControl, "[")
controller.bind(KickLastRacistControl, "]")
controller.bind(KickLastSuspectControl, "\\")
controller.bind(KicksPopControl, "KP_HOME")
controller.bind(KicksClearControl, "KP_LEFTARROW")
controller.bind(KicksPopleftControl, "KP_END")
controller.bind(SpamsPopControl, "KP_PGUP")
controller.bind(SpamsClearControl, "KP_RIGHTARROW")
controller.bind(SpamsPopleftControl, "KP_PGDN")
controller.bind(ClearQueuesControl, "KP_5")
controller.bind(PushControl, "KP_DOWNARROW")


def reset_game() -> None:
    """Start new game."""

    logger.success("RESET GAME")

    global users  # noqa
    users = Users()
    users.me = users.my = users[UserKey(str(config.get("player_name")))]
    users.my.team = Team.BLU
    users.my.display_level = "user"
    ChatsControl.clear()
    MsgQueuesControl.clear()


def debugger() -> None:
    """Drop into python debugger."""

    global conlog  # noqa
    assert conlog

    if conlog.is_eof or SingleStepControl.is_stepping:
        curses.reset_shell_mode()
        breakpoint()  # pylint: disable=forgotten-debug-statement
        curses.reset_prog_mode()


def dump() -> None:
    """Dump stuff."""

    global users  # noqa

    logger.success(pformat(users.__dict__))
    for user in users.users_by_username.values():
        logger.success(pformat(user.__dict__))
